"""Test utils and helper functions."""
import inspect
import json
import os
import shutil
from pathlib import Path

from cdktf_cdktf_provider_google.provider import GoogleProvider

from thipster import Engine
from thipster.auth import Google
from thipster.engine import AuthPort
from thipster.parser import ParserFactory
from thipster.repository import LocalRepo
from thipster.terraform import Terraform

LOCAL_REPO = 'tests/resources/e2e/models'
REMOTE_REPO = 'THipster/models'


class MockAuth(AuthPort):
    """Mock auth class."""

    def authenticate(self, app):
        """Authenticate to GCP."""
        GoogleProvider(
            app, 'default_google',
            project='project_id',
            credentials='credentials.token',
        )


def create_dir(dirname: str, files: dict[str, str]):
    """Create a directory with files.

    Parameters
    ----------
    dirname : str
        The name of the directory to create.
    files : dict[str, str]
        A dictionary of file names and contents.
    """
    if not Path(dirname).is_dir():
        Path(dirname).mkdir()

    dirname = Path(dirname).resolve().as_posix()
    for name, content in files.items():
        create_file(name, content, dirname)

    def destroy_files():
        for content in os.listdir(dirname):
            if not (Path(dirname)/content).is_dir():
                (Path(dirname)/content).unlink()
            else:
                shutil.rmtree(Path(dirname)/content)
        Path(dirname).rmdir()

    return destroy_files


def create_file(filename: str, content: str, dirname: str = 'test'):
    """Create a file with content in the designated directory.

    Parameters
    ----------
    filename : str
        The name of the file to create.
    content : str
        The content of the file to create.
    dirname : str
        The name of the directory in which to create the file.
    """
    if not Path(dirname).is_dir():
        Path(dirname).mkdir()
    dirname = Path(dirname).resolve().as_posix()

    with Path(f'{dirname}/{filename}').open('w') as f:
        f.write(content)


def process_file(
        directory: str, file: str,
        local_repo: str = LOCAL_REPO, file_type: str = 'thips',
        mock_auth=False,
):
    """Handle the file creation, engine run and clean up for the test.

    Parameters
    ----------
    directory : str
        The name of the directory to create.
    file : str
        The content of the file to create.
    local_repo : str
        The path to the local repository.
    file_type : str
        The type of file to create (thips or yaml). Defaults to 'thips'.
    mock_auth : bool
        Whether to mock the authentication to GCP. Defaults to False.
    """
    if not Path('test').is_dir():
        Path('test').mkdir()

    __destroy_dir = create_dir(
        f'test/{directory}',
        {
            f'test_file.{file_type}':
            file,
        },
    )

    def clean_up():
        __destroy_dir()

        if Path('cdktf.out').exists():
            shutil.rmtree('cdktf.out')

        if Path('test').exists() and len(os.listdir('test')) == 0:
            shutil.rmtree('test')

    engine = Engine(
        ParserFactory(),
        LocalRepo(local_repo),
        Google() if not mock_auth else MockAuth(),
        Terraform(),
    )
    try:
        engine.run(f'test/{directory}')
        shutil.move(
            Path(Path.cwd(), 'thipster.tf.json'),
            Path(Path.cwd(), f'test/{directory}', 'thipster.tf.json'),
        )
    except Exception as e:
        clean_up()
        raise e

    return clean_up


def __get_output(test_name):
    with Path(f'test/{test_name}/thipster.tf.json').open() as f:
        file_contents = json.load(f)
        f.close()
    return file_contents


def __get_resource(output_path: str, resource_data: tuple):
    output = __get_output(output_path)

    resources_of_type = output.get('resource').get(resource_data[0])
    resource = None
    for r in resources_of_type.values():
        if r.get('name') == resource_data[1]:
            resource = r

    assert resource is not None

    return resource


def assert_resource_created(
    resource_type: str,
    resource_name: str,
):
    """Assert that a resource was created.

    Parameters
    ----------
    resource_type : str
        The type of resource to check for.
    resource_name : str
        The name of the resource to check for.
    """
    output = __get_output(inspect.currentframe().f_back.f_code.co_name)
    assert output.get('resource') is not None
    resources = output.get('resource')

    assert resources.get(resource_type) is not None

    names = [x.get('name') for _, x in resources.get(resource_type).items()]
    assert resource_name in names

    return (resource_type, resource_name)


def assert_number_of_resource_type_is(
    resource_type: str,
    amount: str,
):
    """Assert that a type of resource was created the right number of times.

    Parameters
    ----------
    resource_type : str
        The type of resource to check for.
    amount : str
        The number of resources to check for.
    """
    output = __get_output(inspect.currentframe().f_back.f_code.co_name)
    assert output.get('resource') is not None
    resources = output.get('resource')

    assert resources.get(resource_type) is not None
    assert len(resources.get(resource_type)) == amount


def assert_resource_parameters_are(resource_data: tuple, parameters: list[str]):
    """Assert that a resource has the right parameters.

    Parameters
    ----------
    resource_data : tuple
        The type and name of the resource to check for.
    parameters : list[str]
        The parameters to check for.
    """
    resource = __get_resource(
        inspect.currentframe().f_back.f_code.co_name, resource_data,
    )

    for parameter in parameters:
        assert parameter in resource


def get_resource_parameter(resource_data: tuple, parameter: str):
    """Get a resource parameter.

    Parameters
    ----------
    resource_data : tuple
        The type and name of the resource to check for.
    parameter : str
        The parameter to get.
    """
    resource = __get_resource(
        inspect.currentframe().f_back.f_code.co_name, resource_data,
    )

    assert parameter in resource
    return resource.get(parameter)


def get_function_name():
    """Get the name of the function that called this function."""
    return inspect.currentframe().f_back.f_code.co_name



"""Exceptions for the THipster engine."""
from abc import ABC, abstractmethod


class THipsterError(Exception, ABC):
    """Base class for all exceptions raised by THipster."""

    @property
    @abstractmethod
    def message():
        """Get the message of the exception."""
        raise NotImplementedError

    def __str__(self) -> str:
        """Get the string representation of the exception."""
        return self.message



"""Module to represent the state and position of the lexer."""
from thipster.engine.parsed_file import Position


class LexerPosition():
    """Represents the state and position of the DSL lexer."""

    def __init__(
        self,
        filename: str = '',
    ) -> None:
        """Represent the state and position of the DSL lexer.

        Parameters
        ----------
        filename : str
            Name of the current file
        """
        self.currentFile = filename
        self.currentLine = 1
        self.currentColumn = 1
        self.currentToken = ''
        self.currentTokenIndex = 1
        self.__currentTokenLine = 1
        self.isVariable = False
        self.isQuotedString = False
        self.isMultiLine = False
        self.isComment = False
        self.escaped = False
        self.consecutiveWhitespaces = 0

    @property
    def currentCharPosition(self) -> Position:  # noqa: N802
        """Get the current position of the lexer."""
        return Position(
            self.currentFile,
            self.currentLine,
            self.currentColumn,
        )

    @property
    def currentTokenPosition(self) -> Position:  # noqa: N802
        """Get the position of the current stored token."""
        return Position(
            self.currentFile,
            self.__currentTokenLine,
            self.currentTokenIndex,
        )

    def new_line(self) -> None:
        """Get the position to the next line."""
        self.currentLine += 1
        self.currentColumn = 1

    def next_column(self, step: int = 1) -> None:
        """Get the position to a new column.

        Parameters
        ----------
        step : int
            Step to modify the current column, by default 1
        """
        if (self.currentColumn + step) > 0:
            self.currentColumn += step
        else:
            self.currentColumn = 0  # Raise exception ?

    def add_to_current_token(self, char) -> None:
        """Add a char to the current stored token.

        Parameters
        ----------
        char : _type_
            Char to add to the token
        """
        self.currentToken += char

    def set_current_token_index(self, new_index: int | None = None) -> None:
        """Modify the stored token index.

        Parameters
        ----------
        new_index : int, optional
            New index of the stored token, by default None
        """
        if new_index:
            self.currentTokenIndex = new_index
        else:
            self.currentTokenIndex = self.currentColumn
            self.__currentTokenLine = self.currentLine

    def reset_current_token(self, new_index: int | None = None) -> None:
        """Reset the current stored token and its index.

        Parameters
        ----------
        new_index : int, optional
            New index of the stored token, by default None
        """
        self.currentToken = ''
        self.set_current_token_index(new_index)

    def increment_consecutive_whitespaces(self) -> None:
        """Add a consecutive whitespace to the lexer state."""
        self.consecutiveWhitespaces += 1

    def reset_consecutive_whitespaces(self) -> None:
        """Reset consecutive whitespaces of the lexer state to 0."""
        self.consecutiveWhitespaces = 0



"""Tests for the DSLParser class."""
from pathlib import Path

import pytest

import thipster.engine.parsed_file as pf
from tests.test_tools import create_dir
from thipster.parser import DSLParser
from thipster.parser.dsl_parser.exceptions import (
    DSLArithmeticError,
    DSLConditionError,
    DSLParserPathNotFoundError,
    DSLSyntaxError,
    DSLUnexpectedEOFError,
)
from thipster.parser.dsl_parser.token import TOKENTYPES as TT


def test_get_files():
    """Test the getfiles method."""
    path_input = 'test'
    _destroy_dir = create_dir(
        path_input,
        {f'test_file{i}.thips': 'content' for i in range(1, 4)},
    )

    parser = DSLParser

    files = parser._DSLParser__getfiles(path_input)

    assert isinstance(files, dict)
    assert len(files) == 3

    for k, v in files.items():
        assert 'test_file' in k
        assert v == 'content'

    _destroy_dir()


def test_get_absent_files():
    """Test the getfiles method with an inexistant path."""
    with pytest.raises(DSLParserPathNotFoundError):

        parser = DSLParser
        parser._DSLParser__getfiles('inexistant_path')


def __test_file(file: str):
    path_input = 'test'
    _destroy_dir = create_dir(
        path_input,
        {
            'test_file.thips':
            file,
        },
    )

    parser = DSLParser
    try:
        output = parser.run(path_input)

        assert isinstance(output, pf.ParsedFile)
    except Exception as e:
        raise e
    finally:
        _destroy_dir()

    return output


def test_parse_simple_file():
    """Test the parsing of a simple file."""
    out = __test_file(
        file="""bucket my-bucket:
\tregion: euw
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert isinstance(region._ParsedAttribute__value, pf.ParsedLiteral)
    assert region.value == 'euw'


def test_parse_empty_file():
    """Test the parsing of an empty file."""
    out = __test_file(
        file="""""",
    )

    assert len(out.resources) == 0


def test_parse_simple_file_with_newlines():
    """Test the parsing of a simple file with newlines."""
    out = __test_file(
        file="""

bucket my-bucket:


\tregion: euw

bucket my-bucket2:
\tregion: euw



""",
    )
    assert len(out.resources) == 2
    for bucket in out.resources:
        assert bucket.resource_type == 'bucket'
        assert 'my-bucket' in bucket.name
        assert len(bucket.attributes) == 1

        region = bucket.attributes[0]
        assert region.name == 'region'
        assert region.value == 'euw'


def test_parse_simple_file_with_empty_lines():
    """Test the parsing of a simple file with empty lines."""
    out = __test_file(
        file="""bucket my-bucket:
\tregion: euw
\t
bucket my-bucket2:
\tregion: euw
\t\t

""",
    )
    assert len(out.resources) == 2
    for bucket in out.resources:
        assert bucket.resource_type == 'bucket'
        assert 'my-bucket' in bucket.name
        assert len(bucket.attributes) == 1

        region = bucket.attributes[0]
        assert region.name == 'region'
        assert region.value == 'euw'


def test_parse_list():
    """Test the parsing of a list."""
    out = __test_file(
        file="""bucket my-bucket:
\tregion:
\t\t- toto
\t\t- titi
\t\t- tata
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert isinstance(region._ParsedAttribute__value, pf.ParsedList)
    assert len(region.value) == 3

    out = __test_file(
        file="""bucket my-bucket:
\tregion: [toto, titi, tata]
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert isinstance(region._ParsedAttribute__value, pf.ParsedList)
    assert len(region.value) == 3

    out = __test_file(
        file="""bucket my-bucket:
\tregion: []
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert isinstance(region._ParsedAttribute__value, pf.ParsedList)
    assert len(region.value) == 0


def test_parse_dict_list_in_dict():
    """Test the parsing of a list and dict in a dict."""
    out = __test_file(
        file="""bucket my-bucket:
\ttoto:
\t\t aaa: val1
\t\t bbb: val2
\ttata:
\t\t- ccc
\t\t- ddd
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 2

    toto = bucket.attributes[0]
    assert toto.name == 'toto'
    assert isinstance(toto._ParsedAttribute__value, pf.ParsedDict)
    tata = bucket.attributes[1]
    assert tata.name == 'tata'
    assert isinstance(tata._ParsedAttribute__value, pf.ParsedList)


def test_parse_if_else():
    """Test the parsing of if else statements."""
    # IN DICT
    out = __test_file(
        file="""bucket my-bucket:
\tregion: euw if 1 == 1 else us
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert region.value == 'euw'

    # ELSE CASE
    out = __test_file(
        file="""bucket my-bucket:
\tregion: euw if 1 == 2 else us
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert region.value == 'us'

    # IN LIST
    out = __test_file(
        file="""bucket my-bucket:
\tregion:
\t\t- euw if 1 == 1 else us
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert isinstance(region._ParsedAttribute__value, pf.ParsedList)

    assert len(region.value) == 1
    r = region.value[0]
    assert r.value == 'euw'


def test_parse_literal_types():
    """Test the parsing of literal types."""
    out = __test_file(
        file="""bucket my-bucket:
\tregion: euw
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert region.value == 'euw'

    out = __test_file(
        file="""bucket my-bucket:
\tregion: 1
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert region.value == 1

    out = __test_file(
        file="""bucket my-bucket:
\tregion: 3.2
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert region.value == 3.2

    out = __test_file(
        file="""bucket my-bucket:
\tregion: true
""",
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert region.value is True


def test_parse_comment():
    """Test the parsing of an amount."""
    out = __test_file(
        file="""bucket my-bucket:
\ttoto: aaaaa // MY COMMENT
\ttata: bbbbb
""",
    )

    assert len(out.resources) == 1
    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert 'my-bucket' in bucket.name
    assert len(bucket.attributes) == 2

    region = bucket.attributes[0]
    assert region.name == 'toto'
    assert region.value == 'aaaaa'
    region = bucket.attributes[1]
    assert region.name == 'tata'
    assert region.value == 'bbbbb'


def test_parse_amount():
    """Test the parsing of an amount."""
    out = __test_file(
        file="""bucket my-bucket: amount: 3
\tregion: euw
""",
    )

    assert len(out.resources) == 3
    for bucket in out.resources:
        assert bucket.resource_type == 'bucket'
        assert 'my-bucket' in bucket.name
        assert len(bucket.attributes) == 1

        region = bucket.attributes[0]
        assert region.name == 'region'
        assert region.value == 'euw'

    out = __test_file(
        file="""bucket my-bucket: amount: 3 #i
\tregion: #i
""",
    )

    assert len(out.resources) == 3
    for bucket, i in zip(out.resources, range(1, 4)):
        assert bucket.resource_type == 'bucket'
        assert 'my-bucket' in bucket.name
        assert len(bucket.attributes) == 1

        region = bucket.attributes[0]
        assert region.name == 'region'
        assert region.value == i


def test_parse_amount_if_value():
    """Test the parsing of an amount."""
    out = __test_file(
        file="""bucket my-bucket: amount: 3 if 1==1 else 2
\tregion: euw
""",
    )

    assert len(out.resources) == 3

    out = __test_file(
        file="""bucket my-bucket: amount: 3 if 1==2 else 2
\tregion: euw
""",
    )

    assert len(out.resources) == 2


def test_parse_if_and_amount_resource():
    """Test the parsing of an amount."""
    out = __test_file(
        file="""bucket my-bucket:  if 1==1 amount: 3
\tregion: euw
""",
    )

    assert len(out.resources) == 3

    out = __test_file(
        file="""bucket my-bucket: if 1==2 amount: 3
\tregion: euw
""",
    )

    assert len(out.resources) == 0


def test_var_in_name():
    """Test the parsing of a variable in a name."""
    out = __test_file(
        file="""bucket my-bucket#i: amount: 3 #i
\tregion: #i
""",
    )

    assert len(out.resources) == 3
    for bucket, i in zip(out.resources, range(1, 4)):
        assert bucket.resource_type == 'bucket'
        assert bucket.name == 'my-bucket' + str(i)
        assert len(bucket.attributes) == 1

        region = bucket.attributes[0]
        assert region.name == 'region'
        assert region.value == i


def test_comparisons():
    """Test the parsing of operator comparisons."""
    def test_cmp(cmp_expr: str, result: str):
        out = __test_file(
            file=f"""bucket my-bucket:
\tregion: truCase if {cmp_expr} else falsCase
""",
        )

        assert len(out.resources) == 1

        bucket = out.resources[0]
        assert bucket.resource_type == 'bucket'
        assert bucket.name == 'my-bucket'
        assert len(bucket.attributes) == 1

        region = bucket.attributes[0]
        assert region.name == 'region'
        assert region.value == result

    # OR
    test_cmp('(1==1 OR 1==2)', 'truCase')

    # AND
    test_cmp('(not 1==1 AND false)', 'falsCase')

    # LTE
    test_cmp('2<=2', 'truCase')
    test_cmp('3<=2', 'falsCase')

    # GTE
    test_cmp('2>=2', 'truCase')
    test_cmp('1>=2', 'falsCase')

    # LT
    test_cmp('1<2', 'truCase')
    test_cmp('2<2', 'falsCase')

    # GT
    test_cmp('3>2', 'truCase')
    test_cmp('2>2', 'falsCase')

    # NE
    test_cmp('3!=2', 'truCase')
    test_cmp('2!=2', 'falsCase')


def test_arithmetic():
    """Test the parsing of arithmetic expressions."""
    def arithmetic_test(calcul: str, expected: int):
        out = __test_file(
            file=f"""bucket my-bucket:
\tvalue: {calcul}
""",
        )

        assert len(out.resources) == 1

        bucket = out.resources[0]
        assert bucket.resource_type == 'bucket'
        assert bucket.name == 'my-bucket'
        assert len(bucket.attributes) == 1

        region = bucket.attributes[0]
        assert region.name == 'value'
        assert region.value == expected

    # PLUS
    arithmetic_test('-1+1-3', 3)

    # OPERATION ORDER
    arithmetic_test('10/2+3*3', 14)
    arithmetic_test('10/(2+3)*3', 6)

    # POW
    arithmetic_test('(-2)^4', 16)

    # MODULO
    arithmetic_test('25%17', 8)


def __test_parser_raises(
    mocker,  # noqa: ARG001
    input_file: str,
    exception: Exception,
) -> pytest.ExceptionInfo:
    with pytest.raises(exception) as exc_info:
        __test_file(
            file=input_file,
        )

    return exc_info


def __test_syntax_error(
    mocker, input_file: str, ln: int, col: int,
    expected: str, got: str,
):

    exc_info = __test_parser_raises(
        mocker,
        input_file,
        DSLSyntaxError,
    )

    exp = str(' or '.join(list(map(str, expected))))\
        if isinstance(expected, list) else str(expected.value)

    assert str(exc_info.value) == f'(File : \
{Path.cwd()}/test/test_file.thips, Ln {ln!s}, Col {col!s}) :\n\t\
Syntax error : Expected {exp}, got {got.value!s}'


def test_syntax_error_resource(mocker):
    """Test the syntax errors in a resource missing its name."""
    input_file = """bucket :
            """
    __test_syntax_error(
        mocker, input_file=input_file, ln=1, col=8,
        expected=[TT.STRING, TT.VAR], got=TT.COLON,
    )


def test_syntax_error_dict(mocker):
    """Test the syntax errors associated with a dict."""
    # MISSING COLON
    input_file = """
bucket my-bucket:
\tregion euw
            """
    __test_syntax_error(
        mocker, input_file=input_file, ln=3, col=9,
        expected=TT.COLON, got=TT.STRING,
    )

    # MISSING TAB
    input_file = """
bucket my-bucket:
region: euw
            """
    __test_syntax_error(
        mocker, input_file=input_file, ln=3, col=1,
        expected=TT.TAB, got=TT.STRING,
    )

    # MISSING VALUE
    input_file = """
bucket my-bucket:
\ttest:
\tregion: euw
"""
    __test_syntax_error(
        mocker, input_file=input_file, ln=3, col=7,
        expected=TT.STRING, got=TT.NEWLINE,
    )

    # MISSING VALUE 2
    input_file = """
bucket my-bucket:
\ttest:





"""
    exc_info = __test_parser_raises(
        mocker, input_file=input_file, exception=DSLUnexpectedEOFError,
    )

    assert str(exc_info.value) == 'Unexpected EOF'


def test_syntax_error_amount(mocker):
    """Test the syntax errors in an amount declaration."""
    # MISSING COLON
    input_file = """
bucket my-bucket: amount 3
\tregion: euw
"""
    __test_syntax_error(
        mocker, input_file=input_file, ln=2, col=26,
        expected=TT.COLON, got=TT.INT,
    )

    # NO INTEGER
    input_file = """
bucket my-bucket: amount: str
\tregion: euw
"""
    __test_parser_raises(
        mocker,
        input_file=input_file,
        exception=DSLArithmeticError,
    )


def test_syntax_error_if(mocker):
    """Test the syntax errors in an if statement."""
    # MISSING CONDITION
    input_file = """
bucket my-bucket: if
"""
    __test_parser_raises(
        mocker,
        input_file=input_file,
        exception=DSLConditionError,
    )

    # UNEXPECTED IF ELSE
    input_file = """
bucket my-bucket: if 1 == 1 else bbb
"""
    __test_syntax_error(
        mocker, input_file=input_file, ln=2, col=29,
        expected=TT.NEWLINE, got=TT.ELSE,
    )


def test_syntax_error_if_else(mocker):
    """Test the syntax errors in an if else statement."""
    # MISSING ELSE VALUE
    input_file = """
bucket my-bucket:
\ttoto:
\t\tfoo : bar if true else
"""
    __test_syntax_error(
        mocker, input_file=input_file, ln=4, col=25,
        expected=[TT.INT, TT.FLOAT, TT.PARENTHESES_START], got=TT.NEWLINE,
    )

    # MISSING CONDITION VALUE
    input_file = """
bucket my-bucket:
\ttoto:
\t\tfoo : bar if  else bbb
"""
    __test_parser_raises(
        mocker,
        input_file=input_file,
        exception=DSLConditionError,
    )


def test_syntax_error_unexpected_token(mocker):
    """Test the syntax error with an unexpected 'if' as token name."""
    # RESERVED TOKEN NAME
    input_file = """
bucket if:
\tregion: euw
"""
    __test_syntax_error(
        mocker, input_file=input_file, ln=2, col=8,
        expected=[TT.STRING, TT.VAR], got=TT.IF,
    )


def test_amount_error(mocker):
    """Test the syntax error with an unexpected 'amount' value : 3/2 isn't an int."""
    # RESERVED TOKEN NAME
    input_file = """
bucket my-bucket: amount: 3/2
\tregion: euw
"""
    __test_parser_raises(
        mocker,
        input_file=input_file,
        exception=DSLArithmeticError,
    )



"""Dedicated module for Terraform-related actions.

This module contains all the code linked to Terraform, including the
generation of the Terraform CDK constructs and the creation of the
Terraform code itself.
It also allows user to execute Terraform commands from the THipster package.
"""

from .cdk import CDK as Terraform



"""Test the YAMLParser class."""
import pytest

import thipster.engine.parsed_file as pf
from tests.test_tools import create_dir
from thipster.parser import YAMLParser
from thipster.parser.yaml_parser import YAMLParserNoNameError


def __test_parser_raises(
    mocker,  # noqa ARG01
    input_file: str,
    exception: Exception,
) -> pytest.ExceptionInfo:
    with pytest.raises(exception) as exc_info:
        __test_file(
            file=input_file,
        )

    return exc_info


def __test_file(file: str):
    path_input = 'test'
    _destroy_dir = create_dir(
        path_input,
        {
            'test_file.yaml':
            file,
        },
    )

    parser = YAMLParser
    try:
        output = parser.run(path_input)

        assert isinstance(output, pf.ParsedFile)
    except Exception as e:
        raise e
    finally:
        _destroy_dir()

    return output


def test_parse_simple_file():
    """Test the parser with a simple yaml file."""
    out = __test_file(
        file="""bucket:
  name: my-bucket
  region: euw
""",
    )
    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert isinstance(region._ParsedAttribute__value, pf.ParsedLiteral)
    assert region.value == 'euw'


def test_parse_simple_jinja_file():
    """Test the parser with a simple file with jinja variables."""
    out = __test_file(
        file="""
{% set name = "test_name" %}
{% set region = "europe-west1" %}
bucket:
  name: {{ name|lower }}
  region: {{ region }}
""",
    )
    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'test_name'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert isinstance(region._ParsedAttribute__value, pf.ParsedLiteral)
    assert region.value == 'europe-west1'


def test_parse_two_resources():
    """Test the parser with a list of two resources."""
    out = __test_file(
        file="""bucket:
  - name: my-bucket1
    region: euw
  - name: my-bucket2
    region: euw
""",
    )
    assert len(out.resources) == 2
    for bucket in out.resources:
        assert bucket.resource_type == 'bucket'
        assert 'my-bucket' in bucket.name
        assert len(bucket.attributes) == 1

        region = bucket.attributes[0]
        assert region.name == 'region'
        assert region.value == 'euw'


def test_parse_dict_in_dict():
    """Test the parser with a dict in a dict."""
    out = __test_file(
        file="""bucket:
  name: my-bucket
  region: euw
  toto:
    aaa: val1
    bbb: val2
""",
    )
    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 2

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert region.value == 'euw'

    toto = bucket.attributes[1]
    assert toto.name == 'toto'
    assert isinstance(toto._ParsedAttribute__value, pf.ParsedDict)


def test_parse_list():
    """Test the parser with a list."""
    out = __test_file(
        file="""bucket:
  name: my-bucket
  toto:
    - aaa
    - bbb
""",
    )
    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    toto = bucket.attributes[0]
    assert toto.name == 'toto'
    assert isinstance(toto._ParsedAttribute__value, pf.ParsedList)
    assert len(toto.value) == 2

    for val in toto.value:
        assert isinstance(val, pf.ParsedLiteral)


def test_parse_dict_in_list():
    """Test the parser with a dict in a list."""
    out = __test_file(
        file="""bucket:
  name: my-bucket
  toto:
    - aaa : 12
      bbb : OK
    - ccc : 10
      ddd : OK
""",
    )
    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    toto = bucket.attributes[0]
    assert toto.name == 'toto'
    assert isinstance(toto._ParsedAttribute__value, pf.ParsedList)
    assert len(toto.value) == 2

    for val in toto.value:
        assert isinstance(val, pf.ParsedDict)
        assert len(val.value) == 2
        for v in val.value:
            assert isinstance(v, pf.ParsedAttribute)


def test_syntax_error_no_name(mocker):
    """Test the parser with a missing resource name."""
    input_file = """bucket:
  toto:
    - aaa
    - bbb
"""
    __test_parser_raises(mocker, input_file, YAMLParserNoNameError)



"""Authentification module interface."""
from abc import ABC, abstractmethod


class AuthPort(ABC):
    """Authentification port."""

    @abstractmethod
    def authenticate(self):
        """Abstract method used for authentication.

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        msg = 'Should implement authenticate()'
        raise NotImplementedError(msg)



"""Test the generation of Terraform files."""
import uuid

import pytest

from tests.test_tools import (
    assert_number_of_resource_type_is,
    assert_resource_created,
    get_function_name,
    get_resource_parameter,
    process_file,
)
from thipster.terraform.exceptions import (
    CDKCyclicDependenciesError,
    CDKDependencyNotDeclaredError,
    CDKMissingAttributeInDependencyError,
)


def test_empty_bucket():
    """Test the generation of an empty bucket."""
    function_name = get_function_name()

    bucket_name = f'empty-bucket-{uuid.uuid4().int}'
    clean_up = process_file(
        directory=function_name,
        file=f"""
bucket {bucket_name}:

""",
        mock_auth=True,
    )

    # Assertions on plan
    assert_number_of_resource_type_is('google_storage_bucket', 1)
    assert_resource_created('google_storage_bucket', bucket_name)

    clean_up()


def test_empty_bucket_two():
    """Test the generation of two buckets (one is empty)."""
    function_name = get_function_name()

    empty_bucket_name = f'empty-bucket-{uuid.uuid4().int}'
    bucket_name = f'empty-bucket-{uuid.uuid4().int}'
    clean_up = process_file(
        directory=function_name,
        file=f"""
bucket {empty_bucket_name}:

bucket {bucket_name}:
    region: europe-west2
""",
        mock_auth=True,
    )

    # Assertions on plan
    assert_number_of_resource_type_is('google_storage_bucket', 2)
    assert_resource_created('google_storage_bucket', empty_bucket_name)
    assert_resource_created('google_storage_bucket', bucket_name)

    clean_up()


def test_dep_with_no_options():
    """Test the generation of a bucket with a missing attribute in its dependency."""
    function_name = get_function_name()

    with pytest.raises(CDKMissingAttributeInDependencyError):
        clean_up = process_file(
            directory=function_name,
            file="""
bucket_bad_dep_parent my-bucket:
\tregion : euw
""",
            mock_auth=True,
        )

        clean_up()


def test_cyclic_deps():
    """Test the generation of a bucket with a cyclic dependency."""
    function_name = get_function_name()

    with pytest.raises(CDKCyclicDependenciesError):
        clean_up = process_file(
            directory=function_name,
            file="""
bucket_bad_dep_cyclic my-bucket:
\tregion : euw
""",
            mock_auth=True,
        )

        clean_up()


def test_default_internal_object():
    """Test the generation of a firewall with default internal objects."""
    function_name = get_function_name()

    clean_up = process_file(
        directory=function_name,
        file="""
firewall testparent:
\tdirection: EGRESS
""",
        mock_auth=True,
    )

    clean_up()


def test_explicit_internal_object():
    """Test the generation of a firewall with explicit internal objects."""
    function_name = get_function_name()
    clean_up = process_file(
        directory=function_name,
        file="""
firewall testparent:
\tdirection: EGRESS

\tallow:
\t\tprotocol: tcp
""",
        mock_auth=True,
    )

    clean_up()


def test_missing_explicit_dependency():
    """Test the generation of a subnetwork with a missing explicit dependency."""
    function_name = get_function_name()

    with pytest.raises(CDKDependencyNotDeclaredError):
        clean_up = process_file(
            directory=function_name,
            file="""
subnetwork lb-subnet:
\tnetwork: lb-net
\tregion: europe-west1b
\tip_range: 10.0.1.0/24
""",
            mock_auth=True,
        )

        clean_up()


def test_explicit_dependency():
    """Test the generation of a subnetwork with an explicit dependency."""
    function_name = get_function_name()

    clean_up = process_file(
        directory=function_name,
        file="""
network lb-net:

subnetwork lb-subnet:
\tnetwork: lb-net
\tregion: europe-west1
\tip_range: 10.0.1.0/24
""",
        mock_auth=True,
    )

    clean_up()


def test_bucket_cors():
    """Test the generation of a bucket with cors."""
    function_name = get_function_name()

    bucket_name = f'cors-bucket-{uuid.uuid4().int}'
    clean_up = process_file(
        directory=function_name,
        file=f"""
bucket {bucket_name}:
    cors:
        origin:
            - "http://example.com"
        method:
            - "*"
        responseHeader:
            - "*"
        maxAge: 400
""",
        mock_auth=True,
    )

    # Assertions on plan
    assert_number_of_resource_type_is('google_storage_bucket', 1)
    assert_resource_created('google_storage_bucket', bucket_name)

    clean_up()


def test_bucket_two_cors():
    """Test the generation of a bucket with two cors."""
    function_name = get_function_name()

    bucket_name = f'cors-bucket-{uuid.uuid4().int}'
    clean_up = process_file(
        directory=function_name,
        file=f"""
bucket:
  name: {bucket_name}
  cors:
    - origin:
      - "http://example.com"
      method:
      - "*"
      responseHeader:
      - "*"
      maxAge: 400
    - origin:
      - "http://example.com/other"
      method:
      - "*"
      responseHeader:
      - "*"
      maxAge: 600
""",
        file_type='yaml',
        mock_auth=True,
    )

    # Assertions on plan
    assert_number_of_resource_type_is('google_storage_bucket', 1)
    bucket = assert_resource_created('google_storage_bucket', bucket_name)
    cors_block = get_resource_parameter(bucket, 'cors')

    assert len(cors_block) == 2

    clean_up()


def test_subnetwork_in_network():
    """Test the generation of multiple subnetworks in a network."""
    function_name = get_function_name()

    network_name = f'network-{uuid.uuid4().int}'
    clean_up = process_file(
        directory=function_name,
        file=f"""
network:
  name: {network_name}
  auto_create_subnetworks: false
  subnetwork:
    - region: europe-west1
      ip_range: 10.0.1.0/24
    - region: us-west1
      ip_range: 10.0.2.0/24
""",
        file_type='yaml',
        mock_auth=True,
    )

    # Assertions on plan
    assert_number_of_resource_type_is('google_compute_network', 1)
    assert_resource_created('google_compute_network', network_name)

    assert_number_of_resource_type_is('google_compute_subnetwork', 2)

    clean_up()



"""Module that contains the THipster DSL Parser."""
import os
from pathlib import Path

from thipster.engine import ParserPort
from thipster.engine.parsed_file import ParsedFile

from .exceptions import DSLParserBaseError, DSLParserPathNotFoundError
from .interpreter import Interpreter
from .lexer import Lexer
from .token_parser import TokenParser


class DSLParser(ParserPort):
    """Parser for the THipster's DSL."""

    @classmethod
    def __getfiles(cls, path: str) -> dict[str, str]:
        """Recursively get all files in the requested directory and its sudirectories.

        Can be run on a path file aswell.

        Parameters
        ----------
        path: str
            Path to run this function onto

        Returns
        -------
        dict[str, str]
            A dictionary that links a filename to its content, fileName : fileContent
        """
        path = Path(path).resolve().as_posix()

        if not Path(path).exists():
            raise DSLParserPathNotFoundError(path)

        files = {}

        if Path(path).is_dir():
            for content in os.listdir(path):
                files.update(DSLParser.__getfiles(f'{path}/{content}'))

        if Path(path).is_file():
            with Path(path).open() as f:
                files.update({path: f.read()})

                f.close()

        return files

    @classmethod
    def run(cls, path: str) -> ParsedFile:
        """Run the DSLParser.

        Parameters
        ----------
        path: str
            Path to run the parser into

        Returns
        -------
        ParsedFile
            A ParsedFile object with the content of all the files in the input path
        """
        try:
            files = DSLParser.__getfiles(path)
            lexer = Lexer(files)
            token_list = lexer.run()
            parser = TokenParser(token_list)
            ast = parser.run()

            interpreter = Interpreter()
            parsed_file = interpreter.run(ast)

            return parsed_file
        except DSLParserBaseError as e:
            raise e
        except Exception as e:
            raise e



"""Tests for the ParsedFile object."""
import thipster.engine.parsed_file as pf


def test_create_parsed_file():
    """Test creating a ParsedFile object."""
    parsed_file = pf.ParsedFile()

    assert isinstance(parsed_file, pf.ParsedFile)


def test_create_resource():
    """Test creating a ParsedResource object."""
    pos = pf.Position('test_file', 3, 7)
    resource = pf.ParsedResource(
        'test_type', 'test_name',
        attributes=None, position=pos,
    )

    assert isinstance(resource, pf.ParsedResource)
    assert resource.resource_type == 'test_type'
    assert resource.name == 'test_name'
    assert resource.attributes is None
    assert resource.position is pos


def test_create_position():
    """Test creating a Position object."""
    pos = pf.Position('test_file', ln=3, col=7)

    assert isinstance(pos, pf.Position)
    assert pos.ln == 3
    assert pos.col == 7


def test_create_num_attr():
    """Test creating a ParsedAttribute object with a number value."""
    val = pf.ParsedLiteral(3)
    pos = pf.Position('test_file', 3, 7)

    attr = pf.ParsedAttribute('test_attr', value=val, position=pos)

    assert isinstance(attr, pf.ParsedAttribute)
    assert isinstance(attr._ParsedAttribute__value, pf.ParsedLiteral)
    assert attr.value == 3


def test_create_str_attr():
    """Test creating a ParsedAttribute object with a string value."""
    val = pf.ParsedLiteral('test_value')
    pos = pf.Position('test_file', 3, 7)

    attr = pf.ParsedAttribute('test_attr', value=val, position=pos)

    assert isinstance(attr, pf.ParsedAttribute)
    assert isinstance(attr._ParsedAttribute__value, pf.ParsedLiteral)
    assert attr.value == 'test_value'


def test_create_list_str_attr():
    """Test creating a ParsedAttribute object with a list value."""
    val = pf.ParsedList(
        [pf.ParsedLiteral('test_value_' + str(i)) for i in range(3)],
    )
    pos = pf.Position('test_file', 3, 7)

    attr = pf.ParsedAttribute('test_attr', value=val, position=pos)

    assert isinstance(attr, pf.ParsedAttribute)
    assert isinstance(attr._ParsedAttribute__value, pf.ParsedList)

    assert len(attr.value) == 3
    for val in attr.value:
        assert 'test_value' in val.value


def test_create_dict_str_attr():
    """Test creating a ParsedAttribute object with a dict value."""
    val = pf.ParsedDict(
        {pf.ParsedLiteral('test_value_' + str(i)) for i in range(3)},
    )
    pos = pf.Position('test_file', 3, 7)

    attr = pf.ParsedAttribute('test_attr', value=val, position=pos)

    assert isinstance(attr, pf.ParsedAttribute)
    assert isinstance(attr._ParsedAttribute__value, pf.ParsedDict)

    assert len(attr.value) == 3
    for val in attr.value:
        assert 'test_value' in val.value


def test_create_composite_attr():
    """Test creating a ParsedAttribute object with a composite value."""
    attr = [
        pf.ParsedAttribute(
            'attr_num',
            value=pf.ParsedLiteral(3),
            position=pf.Position('test_file', 3, 7),
        ),
        pf.ParsedAttribute(
            'attr_str',
            value=pf.ParsedLiteral('test'),
            position=pf.Position('test_file', 4, 7),
        ),
        pf.ParsedAttribute(
            'attr_list',
            value=pf.ParsedList([
                pf.ParsedLiteral('test_list_1'),
                pf.ParsedLiteral('test_list_2'),
                pf.ParsedLiteral('test_list_3'),
            ]),
            position=pf.Position('test_file', 5, 7),
        ),
        pf.ParsedAttribute(
            'attr_dict',
            value=pf.ParsedDict([
                pf.ParsedAttribute(
                    'attr_dict_1',
                    value=pf.ParsedLiteral('test'),
                    position=pf.Position('test_file', 11, 7),
                ),
                pf.ParsedAttribute(
                    'attr_dict_2',
                    value=pf.ParsedLiteral('test'),
                    position=pf.Position('test_file', 12, 7),
                ),
            ]),
            position=pf.Position('test_file', 10, 7),
        ),
    ]

    resource = pf.ParsedResource(
        'test_type',
        'test_name',
        position=pf.Position('test_file', 3, 7),
        attributes=attr,
    )

    assert isinstance(resource, pf.ParsedResource)
    assert resource.resource_type == 'test_type'
    assert isinstance(resource.attributes, list)
    for attribute in resource.attributes:
        assert isinstance(attribute, pf.ParsedAttribute)
        assert isinstance(
            attribute._ParsedAttribute__value,
            pf.ParsedValue,
        )
        if isinstance(attribute._ParsedAttribute__value, pf.ParsedDict):
            for val in attribute.value:
                assert isinstance(val, pf.ParsedAttribute)

        elif isinstance(attribute._ParsedAttribute__value, pf.ParsedList):
            for val in attribute.value:
                assert isinstance(val, pf.ParsedValue)


def test_add_resource():
    """Test adding a resource to a ParsedFile object."""
    parsed_file = pf.ParsedFile()

    pos = pf.Position('test_file', 3, 7)
    resource = pf.ParsedResource(
        'test_type', 'test_name',
        attributes=None, position=pos,
    )

    parsed_file.resources.append(resource)

    assert resource in parsed_file.resources



"""GithubRepo.py module."""

import json

import requests

from .exceptions import ModelNotFoundError
from .json import JSONRepo


class GithubRepo(JSONRepo):
    """Represents a GitHub resources Repository.

    JSON Models of resources and services offered by supported cloud providers are
    stored in a repository.
    This class is used to access those models if they are located in a GitHub repo.
    """

    def __init__(self, repo: str, branch: str = 'main') -> None:
        """Represent a GitHub resources Repository.

        JSON Models of resources and services offered by supported cloud providers are
        stored in a repository.
        This class is used to access those models if they are located in a GitHub repo.

        Parameters
        ----------
        repo : str
            Name of the repository, for example : 'THipster/models'
        branch : str, optional
            Name of the branch, by default 'main'
        """
        super().__init__()
        self.__repo = repo
        self.__branch = branch

    def get_json(self, name: str) -> str | bytes | bytearray:
        """Get the json file from the GitHub repository.

        Parameters
        ----------
        name : str
            Name of the desired resource

        Returns
        -------
        str | bytes | bytearray
            Content of the JSON file defining the desired resource
        """
        config_response = requests.get(
            f'https://raw.githubusercontent.com/{self.__repo}/{self.__branch}/thipster-config.json',
        )

        if not config_response.ok:
            raise ModelNotFoundError(name)

        config_file: dict = json.loads(config_response.content)
        repo_directory = config_file.get('model_folder')

        response = requests.get(
            f'https://raw.githubusercontent.com/{self.__repo}/{self.__branch}/{repo_directory}/{name}.json',
        )

        if not response.ok:
            raise ModelNotFoundError(name)

        return response.content



"""Provider authentication module.

Allows users to authenticate themselves to the desired supported cloud provider.
Currently, only Google Cloud Platform is available.
"""

from .google import GoogleAuth as Google



"""Terraform module interface."""
from abc import ABC, abstractclassmethod

import thipster.engine.parsed_file as pf
import thipster.engine.resource_model as rm

from .i_auth import AuthPort


class TerraformPort(ABC):
    """Terraform port."""

    @classmethod
    @abstractclassmethod
    def apply(
        cls,
        plan_file_path: str | None = None,  # noqa: ARG003
    ):
        """Apply generated terraform code.

        Parameters
        ----------
        plan_file_path : str, optional
            Path to plan file, by default None

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        msg = 'Should implement apply()'
        raise NotImplementedError(msg)

    @classmethod
    @abstractclassmethod
    def generate(
        cls,
        file: pf.ParsedFile,  # noqa: ARG003
        models: dict[str, rm.ResourceModel],  # noqa: ARG003
        _authenticator: AuthPort,
    ):
        """Generate Terraform code from parsed file and models.

        Parameters
        ----------
        file : pf.ParsedFile
            The ParsedFile object containing the resources defined in the input file
        models : dict[str, rm.ResourceModel]
            The dictionary of models

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        msg = 'Should implement generate()'
        raise NotImplementedError(msg)

    @classmethod
    @abstractclassmethod
    def init(cls):
        """Init Terraform for generated terraform code.

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        msg = 'Should implement generate()'
        raise NotImplementedError(msg)

    @classmethod
    @abstractclassmethod
    def plan(cls):
        """Get plan from generated terraform code.

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        msg = 'Should implement plan()'
        raise NotImplementedError(msg)



"""Google Cloud Platform (GCP) authentication module."""
import google.auth
from cdktf_cdktf_provider_google.provider import GoogleProvider

from thipster.engine import AuthPort


class GoogleAuth(AuthPort):
    """Authenticate to Google Cloud Platform (GCP) projects.

    To use this module, you need to have a GCP account and a project created.
    You also need to have gcloud installed : https://cloud.google.com/sdk/docs/install
    Then login using glcloud : gcloud auth application-default login
    """

    @classmethod
    def authenticate(cls, app):
        """Generate the google provider block for the Terraform CDK.

        Parameters
        ----------
        app: Construct
            CDK Construct where the provider is created

        """
        credentials, project_id = google.auth.default()

        GoogleProvider(
            app, 'default_google',
            project=project_id,
            access_token=credentials.token,
        )



"""End to end tests for the thipster tool.

Starts from a .thips or .yaml file and tests the creation of the resources on GCP.
"""
import os
import random
import uuid
from pathlib import Path

import pytest
import tftest

from .test_tools import (
    assert_number_of_resource_type_is,
    assert_resource_created,
    assert_resource_parameters_are,
    get_function_name,
    process_file,
)

AUTH_FILE_PATH = Path(Path.cwd(), 'tests/credentials.json')


@pytest.fixture
def apply_output():
    """Apply terraform and return output."""
    def _apply_output(function_name):
        tf = tftest.TerraformTest(
            tfdir='.',
            basedir=f'test/{function_name}',
        )

        tf.setup()
        try:
            tf.apply()
            yield tf.output()
        except Exception as e:
            raise e

        finally:
            tf.destroy()

    return _apply_output


@pytest.fixture
def authentication():
    """Authenticate to GCP."""
    delete_credentials = False
    if (
        not Path(
            os.getenv('HOME'),
            '.config/gcloud/application_default_credentials.json',
        ).exists()
        and (
            os.getenv('GOOGLE_APPLICATION_CREDENTIALS') is not None
            or os.getenv('GOOGLE_APPLICATION_CREDENTIALS') != ''
        )
    ):

        delete_credentials = True
        if os.getenv('GOOGLE_APPLICATION_CREDENTIALS_CONTENT') is None:
            msg = 'No credentials available'
            raise Exception(msg)

        with Path(AUTH_FILE_PATH).open('w') as auth_file:
            auth_file.write(
                os.environ['GOOGLE_APPLICATION_CREDENTIALS_CONTENT'],
            )
        os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = AUTH_FILE_PATH.as_posix()

    yield

    if delete_credentials:
        Path(AUTH_FILE_PATH).unlink()


def test_bucket(apply_output, authentication):
    """Test bucket creation."""
    _ = authentication
    function_name = get_function_name()

    bucket_name = f'test-bucket-{uuid.uuid4().int}'
    clean_up = process_file(
        directory=function_name,
        file=f"""
bucket {bucket_name}:
\tregion : europe-west1
    """,
    )

    try:
        # Assertions on plan
        assert_number_of_resource_type_is('google_storage_bucket', 1)
        bucket = assert_resource_created('google_storage_bucket', bucket_name)
        assert_resource_parameters_are(bucket, ['location'])

        # Test apply
        _ = list(apply_output(function_name))

    except Exception as e:
        raise e
    finally:
        clean_up()


def test_lb(apply_output, authentication):
    """Test load balancer creation."""
    _ = authentication
    function_name = get_function_name()

    test_id = random.randint(0, 10000)
    clean_up = process_file(
        directory=function_name,
        file=f"""
network lb-net-{test_id}:

subnetwork lb-subnet-{test_id}:
\tnetwork: lb-net-{test_id}
\tregion: europe-west1
\tip_range: 10.0.1.0/24

loadbalancer my-lb-{test_id}:
\tnetwork: lb-net-{test_id}
\tload_balancing_scheme: EXTERNAL
    """,
    )

    try:
        # Assertions on plan
        assert_number_of_resource_type_is('google_compute_network', 1)
        assert_resource_created('google_compute_network', f'lb-net-{test_id}')

        assert_number_of_resource_type_is('google_compute_subnetwork', 1)
        assert_resource_created(
            'google_compute_subnetwork', f'lb-subnet-{test_id}',
        )

        # Test apply
        _ = list(apply_output(function_name))

    except Exception as e:
        raise e
    finally:
        clean_up()



"""Engine.py module."""

import thipster.engine.parsed_file as pf

from .i_auth import AuthPort
from .i_parser import ParserPort
from .i_repository import RepositoryPort
from .i_terraform import TerraformPort
from .resource_model import ResourceModel

terraform_plan_file = 'thipster.tfplan'


class Engine():
    """THipster's Engine.

    The core of the application, it is used to call and link all
    interfaces together.
    """

    def __init__(
            self, parser: ParserPort,
            repository: RepositoryPort,
            auth: AuthPort,
            terraform:  TerraformPort,
    ):
        """THipster's Engine.

        The core of the application, it is used to call and link all
        interfaces together.

        Parameters
        ----------
        parser : ParserPort
            Instance of a Parser class
        repository : RepositoryPort
            Instance of a Respository class
        auth : AuthPort
            Instance of an Auth class
        terraform : TerraformPort
            Instance of a Terraform class
        """
        self.__parser = parser
        self.__repository = repository
        self.__auth = auth
        self.__terraform = terraform

    @property
    def parser(self):
        """Get the parser."""
        return self.__parser

    @parser.setter
    def parser(self, value):
        if not isinstance(value, ParserPort):
            raise Exception

        self.__parser = value

    @property
    def repository(self):
        """Get the repository."""
        return self.__repository

    @repository.setter
    def repository(self, value):
        if not isinstance(value, RepositoryPort):
            raise Exception

        self.__repository = value

    @property
    def auth(self):
        """Get the authentification module."""
        return self.__auth

    @auth.setter
    def auth(self, value):
        if not isinstance(value, AuthPort):
            raise Exception

        self.__auth = value

    @property
    def terraform(self):
        """Get the Terraform module."""
        return self.__terraform

    @terraform.setter
    def terraform(self, value):
        if not isinstance(value, TerraformPort):
            raise Exception

        self.__terraform = value

    def run(self, path: str) -> tuple[list[str], str]:
        """Return json Terraform files from the input file name.

        Calls the different run methods of the parser, repository,
        auth and terraform modules.
        Transforms the inputed filename into a Cloud architecture plan.

        Parameters
        ----------
        path : str
            The path of the files to be processed

        Returns
        -------
        tuple[list[str], str]
            A tuple made up of the list of directories containing the Terraform json
            files and a string with the results of the Terraform plan
        """
        # Parse file or directory
        parsed_file = self.parse_files(path)

        # Get needed models
        models = self.get_models(parsed_file)

        # Generate Terraform files
        self.generate_tf_files(parsed_file, models)

        self.__terraform.init()

        return self.__terraform.plan(terraform_plan_file)

    def parse_files(self, path: str) -> pf.ParsedFile:
        """Parse the input file or directory.

        Parameters
        ----------
        path : str
            The path of the files to be processed

        Returns
        -------
        pf.ParsedFile
            The ParsedFile object containing the resources defined in the input file
        """
        parsed_file = self.__parser.run(path)
        assert type(parsed_file) == pf.ParsedFile

        return parsed_file

    def get_models(
        self, file: pf.ParsedFile,
    ) -> dict[str, ResourceModel]:
        """Get the models from the repository.

        Parameters
        ----------
        file : pf.ParsedFile
            The ParsedFile object containing the resources defined in the input file

        Returns
        -------
        dict[str, ResourceModel]
            The dictionary of models
        """
        types = [r.resource_type for r in file.resources]
        return self.__repository.get(types)

    def generate_tf_files(
        self, file: pf.ParsedFile, models: dict[str, ResourceModel],
    ) -> None:
        """Generate Terraform files.

        Parameters
        ----------
        file : pf.ParsedFile
            The ParsedFile object containing the resources defined in the input file
        models : dict[str, ResourceModel]
            The dictionary of models
        """
        self.__terraform.generate(file, models, self.__auth)

    def init_terraform(self) -> None:
        """Initialize Terraform."""
        self.__terraform.init()

    def plan_terraform(self, plan_file_path: str = terraform_plan_file) -> str:
        """Plan Terraform.

        Parameters
        ----------
        plan_file_path : str, optional
            The path of the plan file, by default thipster.tfplan

        Returns
        -------
        str
            The results of the Terraform plan
        """
        return self.__terraform.plan(plan_file_path)

    def apply_terraform(self, plan_file_path: str = terraform_plan_file) -> str:
        """Apply Terraform.

        Parameters
        ----------
        plan_file_path : str, optional
            The path of the plan file, by default thipster.tfplan

        Returns
        -------
        str
            The results of the Terraform apply
        """
        return self.__terraform.apply(plan_file_path)



"""Github models repository tests."""
import pytest

from thipster.engine.resource_model import ResourceModel
from thipster.repository import GithubRepo
from thipster.repository.exceptions import ModelNotFoundError

repository = 'THipster/models'
branch = 'main'


def test_get_bucket():
    """Test get bucket model."""
    gcp_bucket = 'bucket'
    resources = [gcp_bucket]

    repo = GithubRepo(repository, branch)

    models = repo.get(resources)
    assert isinstance(models, dict)
    assert len(models) == 2

    assert gcp_bucket in models
    assert isinstance(models[gcp_bucket], ResourceModel)

    bucket = models[gcp_bucket]

    assert len(bucket.attributes) == 1
    assert len(bucket.dependencies) == 0
    assert bucket.resource_type == gcp_bucket


def test_get_vm():
    """Test get vm model and its dependencies."""
    gcp_network = 'network'
    gcp_vm = 'vm'
    resources = [gcp_vm]

    repo = GithubRepo(repository, branch)

    models = repo.get(resources)
    assert isinstance(models, dict)
    assert len(models) == 2

    assert gcp_vm in models
    assert isinstance(models[gcp_vm], ResourceModel)

    vm = models[gcp_vm]

    assert vm.resource_type == gcp_vm
    assert len(vm.attributes) == 2
    assert len(vm.dependencies) == 1

    assert gcp_network in models
    assert isinstance(models[gcp_network], ResourceModel)

    network = models[gcp_network]

    assert network.resource_type == gcp_network
    assert len(network.attributes) == 1
    assert len(network.dependencies) == 0


def test_model_not_found():
    """Test get non-existant model."""
    resources = ['not_an_existing_model']

    repo = GithubRepo(repository)

    with pytest.raises(ModelNotFoundError):
        repo.get(resources)



"""Exceptions for the DSL parser."""
from thipster.engine import THipsterError
from thipster.engine.parsed_file import Position

from .token import TOKENTYPES as TT
from .token import Token


class DSLParserBaseError(THipsterError):
    """Shared base exception for the DSL parser."""

    def __init__(self, *args: object) -> None:
        """Shared base exception for the DSL parser."""
        super().__init__(*args)


class DSLParserPathNotFoundError(DSLParserBaseError):
    """Exception raised when a file or directory is not found."""

    def __init__(self, file: str, *args: object) -> None:
        """Exception raised when a file or directory is not found.

        Parameters
        ----------
        file : str
            The file or directory that was not found.
        """
        super().__init__(*args)
        self.file = file

    @property
    def message(self) -> str:
        """Return the exception message."""
        return f'Path not found : {self.file}'


class DSLSyntaxError(DSLParserBaseError):
    """Exception raised when a syntax error is encountered."""

    def __init__(self, token: Token, expected: TT | list[TT], *args: object) -> None:
        """Exception raised when a syntax error is encountered.

        Parameters
        ----------
        token : Token
            The token where the error occured.
        expected : TT | list[TT]
            The expected token type(s).
        """
        super().__init__(*args)
        self.token = token
        self.expected = expected

    @property
    def message(self) -> str:
        """Return the exception message."""
        if type(self.expected) is TT:
            return f"""{self.token.position!s} :\n\tSyntax error : Expected \
{self.expected.value!s}, got {self.token.token_type!s}"""

        return f"""{self.token.position!s} :\n\tSyntax error : Expected \
{' or '.join(list(map(str, self.expected)))!s}, got {
    self.token.token_type!s}"""


class DSLConditionError(DSLParserBaseError):
    """Exception raised when a condition is not valid."""

    def __init__(self, token: Token, *args: object) -> None:
        """Exception raised when a condition is not valid.

        Parameters
        ----------
        token : Token
            The token where the error occured.
        """
        super().__init__(*args)
        self.token = token

    @property
    def message(self) -> str:
        """Return the exception message."""
        return f'{self.token.position!s} :\n\tBad condition'


class DSLArithmeticError(DSLParserBaseError):
    """Exception raised when an arithmetic error is encountered."""

    def __init__(self, position: Position, error_message: str, *args: object) -> None:
        """Exception raised when an arithmetic error is encountered.

        Parameters
        ----------
        position : Position
            The position where the error occured.
        error_message : str
            The error message.
        """
        super().__init__(*args)
        self.position = position
        self.error_message = error_message

    @property
    def message(self) -> str:
        """Return the exception message."""
        return f'{self.position!s} :\n\tArithmetic error :{self.error_message}'


class DSLUnexpectedEOFError(DSLParserBaseError):
    """Exception raised when an unexpected EOF is encountered."""

    def __init__(self, *args: object) -> None:
        """Exception raised when an unexpected EOF is encountered."""
        super().__init__(*args)

    @property
    def message(self) -> str:
        """Return the exception message."""
        return 'Unexpected EOF'


class DSLParserNoEndingQuotesError(DSLParserBaseError):
    """Exception raised when no ending quotes are found."""

    def __init__(self, position: Position, *args: object) -> None:
        """Exception raised when no ending quotes are found.

        position : Position
            The position where the error occured.
        """
        super().__init__(*args)
        self.position = position

    @property
    def message(self) -> str:
        """Return the exception message."""
        return f'Invalid syntax, missing ending quotes at : {self.position}'


class DSLParserVariableAlreadyUsedError(DSLParserBaseError):
    """Exception raised when a variable name is already used."""

    def __init__(self, var: str, *args: object) -> None:
        """Exception raised when a variable name is already used.

        Parameters
        ----------
        var : str
            The variable name.
        """
        super().__init__(*args)
        self.variable = var

    @property
    def message(self) -> str:
        """Return the exception message."""
        return f'Variable already used : {self.variable}'


class DSLParserVariableNotDeclaredError(DSLParserBaseError):
    """Exception raised when a variable is not declared."""

    def __init__(self, var: str, *args: object) -> None:
        """Exception raised when a variable is not declared.

        Parameters
        ----------
        var : str
            The variable name.
        """
        super().__init__(*args)
        self.variable = var

    @property
    def message(self) -> str:
        """Return the exception message."""
        return f'Variable not declared : {self.variable}'



"""Tests for the parser factory."""
import pytest

from tests.test_tools import create_dir
from thipster.engine.parsed_file import ParsedFile
from thipster.parser import ParserFactory
from thipster.parser.exceptions import NoFileFoundError


def __test_file(files: str):
    path_input = 'test'
    _destroy_dir = create_dir(
        path_input,
        files,
    )

    parser = ParserFactory()
    try:
        output = parser.run(path_input)

        assert type(output) == ParsedFile
    except Exception as e:
        raise e
    finally:
        _destroy_dir()

    return output


def test_yaml_file():
    """Test the parser with .YAML and .YML files."""
    # .YAML
    out = __test_file(
        {
            'test_file.yaml':
            """bucket:
  name: my_bucket
  region: euw
""",
        },
    )
    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my_bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert region.value == 'euw'

    # .YML
    out = __test_file(
        {
            'test_file.yml':
            """bucket:
  name: my-bucket
  region: euw
""",
        },
    )
    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert region.value == 'euw'


def test_thips_file():
    """Test the parser with .THIPS files."""
    out = __test_file(
        {
            'test_file.thips':
            """bucket my-bucket:
\tregion: euw
""",
        },
    )
    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert region.value == 'euw'


def test_two_types_of_file():
    """Test the parser with both .THIPS and .YAML files."""
    out = __test_file(
        {
            'test_file.thips':
            """bucket my-bucket1:
\tregion: euw
""",
            'test_file.yml':
            """bucket:
  name: my-bucket2
  region: euw
""",
        },
    )
    assert len(out.resources) == 2

    for bucket in out.resources:
        assert bucket.resource_type == 'bucket'
        assert 'my-bucket' in bucket.name
        assert len(bucket.attributes) == 1

        region = bucket.attributes[0]
        assert region.name == 'region'
        assert region.value == 'euw'


def test_no_parsable_file():
    """Test the parser with no parsable files."""
    with pytest.raises(NoFileFoundError):
        __test_file(
            {
                'test_file.txt':
                """ """,
                'test_file2.txt':
                """ """,
            },
        )


def test_one_parsable_file():
    """Test the parser with one parsable file and one non-parsable file."""
    out = __test_file(
        {
            'test_file.thips':
            """bucket my-bucket1:
\tregion: euw
""",
            'test_file2.txt':
            """ """,
        },
    )

    assert len(out.resources) == 1

    bucket = out.resources[0]
    assert bucket.resource_type == 'bucket'
    assert bucket.name == 'my-bucket1'
    assert len(bucket.attributes) == 1

    region = bucket.attributes[0]
    assert region.name == 'region'
    assert region.value == 'euw'



"""YAML Parser module."""
import os
from abc import ABC
from pathlib import Path

import yaml
from jinja2 import Environment, FileSystemLoader

import thipster.engine.parsed_file as pf
from thipster.engine import ParserPort, THipsterError

from .exceptions import ParserPathNotFoundError


class YAMLParserBaseError(THipsterError, ABC):
    """Base error for YAMLParser."""

    def __init__(self, *args: object) -> None:
        """Shared base exception for the YAML parser."""
        super().__init__(*args)


class YAMLParserNoNameError(YAMLParserBaseError):
    """Error raised when a resource has no name."""

    def __init__(self, resource, *args: object) -> None:
        """Error raised when a resource has no name.

        Parameters
        ----------
        resource
            The resource that has no name.
        """
        super().__init__(*args)
        self.resource = resource

    @property
    def message(self) -> str:
        """Return the error message."""
        return f'No name for resource : {self.resource}'


class YAMLParser(ParserPort):
    """YAMLParser class, used to parse YAML files."""

    @classmethod
    def __getfiles(cls, path: str) -> list[str]:
        """Get the file(s) on the requested path.

        Recursively get all files names in the requested directory and its
        sudirectories. Can be run on a path file as well.

        Parameters
        ----------
        path: str
            Path to run this function onto

        Returns
        -------
        list[str]
            A list of all the filenames
        """
        path = Path(path).resolve().as_posix()

        if not Path(path).exists():
            raise ParserPathNotFoundError(path)

        files = []

        if Path(path).is_dir():
            for content in os.listdir(path):
                files += cls.__getfiles(f'{path}/{content}')

        if Path(path).is_file():
            return [path]

        return files

    @classmethod
    def run(cls, path: str) -> pf.ParsedFile:
        """Run the YAMLParser.

        Parameters
        ----------
        path: str
            Path to run the parser into

        Returns
        -------
        ParsedFile
            A ParsedFile object with the content of all the files in the input path
        """
        try:
            files = cls.__getfiles(path)
            parsed_file = pf.ParsedFile()

            for file in files:
                filedir, filename = os.path.split(file)

                environment = Environment(
                    loader=FileSystemLoader(filedir),
                    autoescape=True,
                )
                template = environment.get_template(filename)
                rendered = template.render()
                content = yaml.safe_load(rendered)

                parsed_file.resources += cls.__convert(content)

            return parsed_file
        except yaml.YAMLError as exc:
            raise exc
        except YAMLParserBaseError as e:
            raise e
        except Exception as e:
            raise e

    @classmethod
    def __convert(cls, file: dict) -> list[pf.ParsedResource]:
        """Convert a dictionnary into a list of ParsedResources.

        Parameters
        ----------
        file: dict
            Dict to convert

        Returns
        -------
        list[ParsedResource]
            Converted dict
        """
        resources = []

        for key, val in file.items():
            if type(val) == list:
                for res in val:
                    if 'name' not in res:
                        raise YAMLParserNoNameError(key)

                    name = res['name']
                    del res['name']

                    resources.append(
                        cls.__get_resource(
                            content=res, resource_type=key, name=name,
                        ),
                    )
            elif type(val) == dict:
                if 'name' not in val:
                    raise YAMLParserNoNameError(key)

                name = val['name']
                del val['name']

                resources.append(
                    cls.__get_resource(
                        content=val, resource_type=key, name=name,
                    ),
                )

        return resources

    @classmethod
    def __get_resource(cls, content: dict, resource_type: str, name: str)\
            -> pf.ParsedResource:
        """Convert a dict in a ParsedResource.

        Parameters
        ----------
        content: dict
            Dict of attributes of the resource
        resourceType: str
            Type of the resource
        name: str
            Name of the resource

        Returns
        -------
        ParsedResource
            A ParsedResource object with the content of the dict
        """
        attr = []

        for key, val in content.items():
            attr.append(cls.__get__attr(key, val))

        return pf.ParsedResource(
            parsed_resource_type=resource_type,
            name=name,
            position=None,
            attributes=attr,
        )

    @classmethod
    def __get__attr(cls, name: str, value: object) -> pf.ParsedAttribute:
        """Convert an object in a ParsedAttribute.

        Parameters
        ----------
        value: object
            Object to convert
        name: str
            Name of the attribute

        Returns
        -------
        ParsedResource
            A ParsedAttribute object with wanted value type
        """
        if type(value) == dict:
            val = cls.__get_dict(value)
        elif type(value) == list:
            val = cls.__get_list(value)
        else:
            val = pf.ParsedLiteral(value)

        return pf.ParsedAttribute(
            name=name,
            value=val,
            position=None,
        )

    @classmethod
    def __get_dict(cls, input_dict: dict) -> pf.ParsedDict:
        """Convert a dict into a list of ParsedDict.

        Parameters
        ----------
        input_dict: dict
            Dict to convert

        Returns
        -------
        ParsedDict
            Converted dict
        """
        attr = []

        for key, val in input_dict.items():
            attr.append(cls.__get__attr(key, val))

        return pf.ParsedDict(attr)

    @classmethod
    def __get_list(cls, input_list: list) -> pf.ParsedList:
        """Convert a dictionnary into a ParsedList.

        Parameters
        ----------
        input_list: list
            List to convert

        Returns
        -------
        ParsedList
            Converted list
        """
        attr = []

        for val in input_list:
            if isinstance(val, dict):
                attr.append(cls.__get_dict(val))
            else:
                attr.append(pf.ParsedLiteral(val))

        return pf.ParsedList(attr)



"""Interpreter for the DSL Parser."""
import thipster.engine.parsed_file as pf
import thipster.parser.dsl_parser.ast as ast

from .exceptions import (
    DSLArithmeticError,
    DSLParserVariableAlreadyUsedError,
    DSLParserVariableNotDeclaredError,
)
from .token import TOKENTYPES as TT


class Interpreter:
    """Interpreter class for the DSL Parser.

    Implements a visitor design pattern on the AST nodes
    """

    def __init__(self) -> None:
        """Interpreter class for the DSL Parser.

        Implements a visitor design pattern on the AST nodes
        """
        super().__init__()

        self.__variables = dict[str, int]()

    def run(self, tree: ast.FileNode) -> pf.ParsedFile:
        """Run the interpreter on an Abstract Syntax Tree.

        Parameters
        ----------
        tree: FileNode
            The root node of the abstract syntax tree

        Returns
        -------
        ParsedFile
            The parsed file stucture assiociated to the given AST
        """
        return tree.accept(self)

    def visit_comp_expr(self, element: ast.CompExprNode) -> bool:
        """Visitor for a CompExprNode.

        Parameters
        ----------
        element: CompExprNode
            The visited node

        Returns
        -------
        bool
            The boolean value of the node
        """
        match element.operation.token_type:
            case TT.NOT:
                return pf.ParsedLiteral(not element.left_value.accept(self).value)

            case TT.OR:
                return pf.ParsedLiteral(
                    element.left_value.accept(self).value or
                    element.right_value.accept(self).value,
                )

            case TT.AND:
                return pf.ParsedLiteral(
                    element.left_value.accept(self).value and
                    element.right_value.accept(self).value,
                )

            case TT.EE:
                return pf.ParsedLiteral(
                    element.left_value.accept(self).value ==
                    element.right_value.accept(self).value,
                )

            case TT.NE:
                return pf.ParsedLiteral(
                    element.left_value.accept(self).value !=
                    element.right_value.accept(self).value,
                )

            case TT.LT:
                return pf.ParsedLiteral(
                    element.left_value.accept(self).value <
                    element.right_value.accept(self).value,
                )

            case TT.LTE:
                return pf.ParsedLiteral(
                    element.left_value.accept(self).value <=
                    element.right_value.accept(self).value,
                )

            case TT.GT:
                return pf.ParsedLiteral(
                    element.left_value.accept(self).value >
                    element.right_value.accept(self).value,
                )

            case TT.GTE:
                return pf.ParsedLiteral(
                    element.left_value.accept(self).value >=
                    element.right_value.accept(self).value,
                )

    def visit_arith_expr(self, element: ast.ArithExprNode) -> int | float:
        """Visitor for an ArithExprNode.

        Parameters
        ----------
        element: ArithExprNode
            The visited node

        Returns
        -------
        int | float
            The arithmetic value of the node
        """
        total = element.terms[0].accept(self)
        if isinstance(total, pf.ParsedLiteral):
            total = total.value

        for i in range(len(element.operations)):
            match element.operations[i]:
                case TT.PLUS:
                    add = element.terms[i+1].accept(self)
                    if isinstance(add, pf.ParsedLiteral):
                        add = add.value
                    total += add

                case TT.MINUS:
                    rm = element.terms[i+1].accept(self)
                    if isinstance(rm, pf.ParsedLiteral):
                        rm = rm.value
                    total -= rm

        return pf.ParsedLiteral(total)

    def visit_term(self, element: ast.TermNode) -> int | float:
        """Visitor for a TermNode.

        Parameters
        ----------
        element: TermNode
            The visited node

        Returns
        -------
        int | float
            The arithmetic value of the node
        """
        total = element.factors[0].accept(self)
        if isinstance(total, pf.ParsedLiteral):
            total = total.value

        for i in range(len(element.operation)):
            match element.operation[i]:
                case TT.MUL:
                    mul = element.factors[i+1].accept(self)
                    if isinstance(mul, pf.ParsedLiteral):
                        mul = mul.value
                    total *= mul

                case TT.DIV:
                    div = element.factors[i+1].accept(self)
                    if isinstance(div, pf.ParsedLiteral):
                        div = div.value
                    total /= div

                case TT.PERCENT:
                    mod = element.factors[i+1].accept(self)
                    if isinstance(mod, pf.ParsedLiteral):
                        mod = mod.value
                    total %= mod

        return total

    def visit_factor(self, element: ast.FactorNode) -> int | float:
        """Visitor for a Factor node.

        Parameters
        ----------
        element: FactorNode
            The visited node

        Returns
        -------
        int | float
            The arithmetic value of the node
        """
        total = element.factors[0].accept(self)
        if isinstance(total, pf.ParsedLiteral):
            total = total.value

        match element.operation:
            case TT.PLUS:
                return pf.ParsedLiteral(total)

            case TT.MINUS:
                return pf.ParsedLiteral(-total)

            case TT.POW:
                for f in element.factors[1:]:
                    power = f.accept(self)
                    if isinstance(power, pf.ParsedLiteral):
                        power = power.value
                    total = total**power
                return pf.ParsedLiteral(total)

    def visit_string_expr(self, element: ast.StringExprNode) -> str:
        """Visitor for a StringExprNode.

        Parameters
        ----------
        element: StringExprNode
            The visited node

        Returns
        -------
        str
            The string value of the node
        """
        ret = ''
        for value in element.values:
            ret += str(value.accept(self))

        return pf.ParsedLiteral(ret)

    def visit_variable_definition(self, element: ast.VariableDefinitionNode) -> str:
        """Visitor for a VariableDefinitionNode.

        Parameters
        ----------
        element: VariableDefinitionNode
            The visited node

        Returns
        -------
        str
            The name of the declared variable

        Raises
        ------
        DSLParserVariableAlreadyUsed
            Tried to declare a varible that was already declared
        """
        if element.name in self.__variables:
            raise DSLParserVariableAlreadyUsedError(element.name)

        self.__variables[element.name] = element.value.accept(self)

        return element.name

    def visit_variable(self, element: ast.VariableNode) -> object:
        """Visitor for a VariableNode.

        Parameters
        ----------
        element: VariableNode
            The visited node

        Returns
        -------
        object
            The value of the declared variable

        Raises
        ------
        DSLParserVariableNotDeclared
            Tried to use a varible that was not declared
        """
        if element.name not in self.__variables:
            raise DSLParserVariableNotDeclaredError(element.name)

        return self.__variables[element.name]

    def visit_int(self, element: ast.IntNode) -> int:
        """Visitor for an IntNode.

        Parameters
        ----------
        element: IntNode
            The visited node

        Returns
        -------
        int
            The value of the node
        """
        return int(element.token.value)

    def visit_float(self, element: ast.FloatNode) -> float:
        """Visitor for a FloatNode.

        Parameters
        ----------
        element: FloatNode
            The visited node

        Returns
        -------
        float
            The value of the node
        """
        return float(element.token.value)

    def visit_bool(self, element: ast.BoolNode) -> bool:
        """Visitor for a BoolNode.

        Parameters
        ----------
        element: BoolNode
            The visited node

        Returns
        -------
        bool
            The value of the node
        """
        return bool(element.token.value)

    def visit_string(self, element: ast.StringNode) -> str:
        """Visitor for a BoolNode.

        Parameters
        ----------
        element: BoolNode
            The visited node

        Returns
        -------
        bool
            The value of the node
        """
        return str(element.token.value)

    def visit_if(self, element: ast.IfNode) -> object | None:
        """Visitor for an IfNode.

        Parameters
        ----------
        element: IfNode
            The visited node

        Returns
        -------
        object | None
            The value of the child node if the condition is true, else None
        """
        return element.if_case.accept(self) if element.condition.accept(self).value\
            else None

    def visit_ifelse(self, element: ast.IfElseNode):
        """Visitor for an IfElseNode.

        Parameters
        ----------
        element: IfElseNode
            The visited node

        Returns
        -------
        object | None
            The value of the "if" child node if the condition is true, else the value\
            of the "else" child node
        """
        return element.if_case.accept(self) if element.condition.accept(self).value\
            else element.else_case.accept(self)

    def visit_amount(self, element: ast.AmountNode) -> list[object]:
        """Visitor for an AmountNode.

        Parameters
        ----------
        element: AmountNode
            The visited node

        Returns
        -------
        list[object]
            A list with the required number of object (resource, list item, ...)
        """
        var = element.variable.accept(self) if element.variable else None
        res = []
        amount = element.amount.accept(self).value
        if not isinstance(amount, int):
            raise DSLArithmeticError(
                element.amount.position, 'Integer expected',
            )

        for _ in range(amount):
            node = element.node.accept(self)
            if node:
                res += node
            if var:
                self.__variables[var] += 1
        return res

    def visit_parameter(self, element: ast.ParameterNode) -> pf.ParsedAttribute:
        """Visitor for a ParameterNode.

        Parameters
        ----------
        element: ParameterNode
            The visited node

        Returns
        -------
        ParsedAttribute
            A ParsedAttribute object containing the name, position and value of \
                the node
        """
        return pf.ParsedAttribute(
            name=element.name.accept(self),
            position=element.position,
            value=element.value.accept(self),
        )

    def visit_dict(self, element: ast.DictNode) -> pf.ParsedDict:
        """Visitor for a DictNode.

        Parameters
        ----------
        element: DictNode
            The visited node

        Returns
        -------
        ParsedDict
            A ParsedDict object based on the node attributes
        """
        return pf.ParsedDict([v.accept(self) for v in element.values])

    def visit_literal(self, element: ast.LiteralNode) -> pf.ParsedLiteral:
        """Visitor for a LiteralNode.

        Parameters
        ----------
        element: LiteralNode
            The visited node

        Returns
        -------
        ParsedLiteral
            A ParsedLiteral object based on the node value
        """
        return pf.ParsedLiteral(element.value.accept(self))

    def visit_list(self, element: ast.ListNode) -> pf.ParsedList:
        """Visitor for a ListNode.

        Parameters
        ----------
        element: ListNode
            The visited node

        Returns
        -------
        ParsedList
            A ParsedLiteral object based on the node elements
        """
        return pf.ParsedList([v.accept(self) for v in element.values])

    def visit_resource(self, element: ast.ResourceNode) -> list[pf.ParsedResource]:
        """Visitor for a ResourceNode.

        Parameters
        ----------
        element: ResourceNode
            The visited node

        Returns
        -------
        int
            A ParsedResource object based on the node's attributes
        """
        return [
            pf.ParsedResource(
                parsed_resource_type=element.type.accept(self).value,
                name=element.name.accept(self).value,
                position=element.position,
                attributes=[v.accept(self) for v in element.parameters.values],
            ),
        ]

    def visit_file(self, element: ast.FileNode):
        """Visitor for a FileNode.

        Parameters
        ----------
        element: FileNode
            The visited node

        Returns
        -------
        ParsedFile
            A ParsedFile object based on the node's resources
        """
        parsed_file = pf.ParsedFile()
        for res in element.resources:
            parsed_file.resources += res.accept(self)

        return parsed_file



"""THipster's DSL AST (Abstract Syntax Tree) module."""
from abc import ABC, abstractmethod

from thipster.engine.parsed_file import Position

from .token import Token


class Node(ABC):
    """Abstract base AST node."""

    @abstractmethod
    def accept(self, visitor):
        """Accept visitor method."""
        raise NotImplementedError

    @property
    def position(self):
        """Return node's position in the file."""
        raise NotImplementedError


class StringNode(Node):
    """String node."""

    def __init__(self, token: Token) -> None:
        super().__init__()
        self.token = token

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<STRING {self.token}>'

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.token.position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_string(self)


class StringExprNode(Node):
    """String expression node."""

    def __init__(self, *values: Node | list[Node]) -> None:
        super().__init__()
        self.values: list[Node] = []
        for v in values:
            if isinstance(v, list):
                self.values.extend(v)
            else:
                self.values.append(v)

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f"<STRING-EXPR {' '.join(list(map(str, self.values)))}>"

    @property
    def position(self) -> Token:
        """Return node's position in the file."""
        return self.values[0].position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_string_expr(self)


class VariableNode(Node):
    """Variable node."""

    def __init__(self, token: Token) -> None:
        super().__init__()
        self.token = token

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<VAR {self.token}>'

    @property
    def name(self):
        """Return variable's name."""
        return self.token.value

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.token.position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_variable(self)


class BoolNode(Node):
    """Boolean node."""

    def __init__(self, token: Token) -> None:
        super().__init__()
        self.token = token

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<BOOL {self.token}>'

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.token.position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_bool(self)


class IntNode(Node):
    """Integer node."""

    def __init__(self, token: Token) -> None:
        super().__init__()
        self.token = token

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<INT {self.token}>'

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.token.position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_int(self)


class FloatNode(Node):
    """Float node."""

    def __init__(self, token: Token) -> None:
        super().__init__()
        self.token = token

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<FLOAT {self.token}>'

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.token.position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_float(self)


class VariableDefinitionNode(Node):
    """Variable definition node."""

    def __init__(self, name: Token, value: IntNode) -> None:
        super().__init__()
        self.__name = name
        self.value = value

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<VARDEF {self.__name} = {self.value}>'

    @property
    def name(self):
        """Return variable's name."""
        return self.__name.value

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.__name.position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_variable_definition(self)


class IfNode(Node):
    """If node."""

    def __init__(self, condition: StringNode, if_case: Node) -> None:
        super().__init__()
        self.condition = condition
        self.if_case = if_case

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<IF {self.condition}: {self.if_case}>'

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.if_case.position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_if(self)


class IfElseNode(IfNode):
    """If-else node."""

    def __init__(
        self, condition: StringNode, if_case: Node,
        else_case: Node | None,
    ) -> None:
        super().__init__(condition, if_case)
        self.else_case = else_case

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<IF_E {self.condition!s}: {self.if_case!s}, ELSE : \
{self.else_case!s}>'

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_ifelse(self)


class AmountNode(Node):
    """Amount node."""

    def __init__(
        self, position: Token,
        amount: Node, variable: VariableDefinitionNode | None, node: Node | None,
    )\
            -> None:
        super().__init__()
        self.__position = position
        self.amount = amount
        self.node = node
        self.variable = variable

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<AMOUNT {self.amount} #{self.variable}: {self.node}>'

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.__position

    @position.setter
    def position(self, _value) -> Position:
        self.__position = _value

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_amount(self)


class ValueNode(Node, ABC):
    """Abstract base value node."""

    pass


class ParameterNode(Node):
    """Parameter node."""

    def __init__(self, name: StringNode, value: ValueNode) -> None:
        super().__init__()
        self.name = name
        self.value = value

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<PARAMETER name = {self.name!s}, value = {self.value!s}>'

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.name.position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_parameter(self)


class DictNode(ValueNode):
    """Dictionary node."""

    def __init__(self, values: list[ParameterNode]) -> None:
        super().__init__()
        self.values = values

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f"<DICT {' '.join(list(map(str, self.values)))}>"

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.values[0].position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_dict(self)


class LiteralNode(ValueNode):
    """Literal node."""

    def __init__(self, values: Node) -> None:
        super().__init__()
        self.value = values

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<LITERAL {self.value!s}>'

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.value.position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_literal(self)


class ListNode(ValueNode):
    """List node."""

    def __init__(self, values: list[ValueNode]) -> None:
        super().__init__()
        self.values = values

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f"<LIST {' '.join(list(map(str, self.values)))}>"

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.values[0].position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_list(self)


class CompExprNode(Node):
    """Comparison expression node."""

    def __init__(
            self, left_value: Node, operation, right_value: Node | None = None,
    ) -> None:
        super().__init__()
        self.left_value = left_value
        self.operation = operation
        self.right_value = right_value

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.left_value.position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_comp_expr(self)


class TermNode(Node):
    """Term node."""

    def __init__(self, factors, operation) -> None:
        super().__init__()
        if len(factors) != len(operation) + 1:
            raise
        self.factors: list[FactorNode] = factors
        self.operation = operation

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.factors[0].position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_term(self)


class ArithExprNode(Node):
    """Arithmetic expression node."""

    def __init__(self, terms: list[TermNode], operations) -> None:
        super().__init__()
        if len(terms) != len(operations) + 1:
            raise
        self.terms = terms
        self.operations = operations

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.terms[0].position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_arith_expr(self)


class FactorNode(Node):
    """Factor node."""

    def __init__(self, factors: list[Node], operation) -> None:
        super().__init__()
        self.factors = factors
        self.operation = operation

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.factors[0].position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_factor(self)


class ResourceNode(Node):
    """Resource node."""

    def __init__(
        self,
        resource_type: StringNode,
        name: StringNode,
        parameters: DictNode | ListNode,
    ) -> None:
        super().__init__()
        self.type = resource_type
        self.name = name
        self.parameters = parameters

    def __str__(self) -> str:
        """Return string representation of the node."""
        return f'<RESOURCE type = {self.type!s}, \
name = {self.name!s}, parameters = {self.parameters!s}>'

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.type.position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_resource(self)


class FileNode(Node):
    """File node."""

    def __init__(self) -> None:
        super().__init__()
        self.resources: list[ResourceNode | IfNode | AmountNode] = []

    def add(self, item: ResourceNode | IfNode | AmountNode) -> None:
        """Add item to the node."""
        self.resources.append(item)

    def __str__(self) -> str:
        """Return string representation of the node."""
        return '\n'.join(list(map(str, self.resources)))

    @property
    def position(self) -> Position:
        """Return node's position in the file."""
        return self.resources[0].position

    def accept(self, visitor):
        """Accept visitor method."""
        return visitor.visit_file(self)



"""Test the THipster DSL token parser."""
from thipster.engine.parsed_file import Position
from thipster.parser.dsl_parser.token import TOKENTYPES as TT
from thipster.parser.dsl_parser.token import Token
from thipster.parser.dsl_parser.token_parser import TokenParser


def test_simple_file():
    """Test the correct parsing of a simple tokenized file."""
    tokenized_input = [
        Token(Position('file', 1, 1), TT.STRING, 'bucket'),
        Token(Position('file', 1, 7), TT.WHITESPACE),
        Token(Position('file', 1, 8), TT.STRING, 'nom'),
        Token(Position('file', 1, 11), TT.MINUS),
        Token(Position('file', 1, 12), TT.INT, 8),
        Token(Position('file', 1, 13), TT.COLON),
        Token(Position('file', 1, 14), TT.NEWLINE),
        Token(Position('file', 2, 1), TT.TAB),
        Token(Position('file', 2, 2), TT.STRING, 'region'),
        Token(Position('file', 1, 8), TT.COLON),
        Token(Position('file', 2, 3), TT.STRING, 'euw'),
        Token(Position('file', 1, 14), TT.NEWLINE),
        Token(Position('file', 2, 4), TT.EOF),
    ]

    parser = TokenParser(tokenized_input)
    output = parser.run()

    assert str(output) == '<RESOURCE \
type = <STRING-EXPR <STRING (STRING bucket)>>, \
name = <STRING-EXPR <STRING (STRING nom)> <STRING (STRING -)> <INT (INT 8)>>, \
parameters = <DICT <PARAMETER name = <STRING (STRING region)>, \
value = <STRING-EXPR <STRING (STRING euw)>>>>>'


def test_newline_remover():
    """Test the correct parsing of a tokenized file with multiple newlines."""
    tokenized_input = [
        Token(Position('file', 1, 14), TT.NEWLINE),
        Token(Position('file', 1, 14), TT.NEWLINE),
        Token(Position('file', 1, 1), TT.STRING, 'bucket'),
        Token(Position('file', 1, 1), TT.WHITESPACE),
        Token(Position('file', 1, 8), TT.STRING, 'nom'),
        Token(Position('file', 1, 8), TT.MINUS),
        Token(Position('file', 1, 8), TT.INT, 8),
        Token(Position('file', 1, 8), TT.COLON),
        Token(Position('file', 1, 14), TT.NEWLINE),
        Token(Position('file', 1, 14), TT.NEWLINE),
        Token(Position('file', 2, 1), TT.TAB),
        Token(Position('file', 2, 2), TT.STRING, 'region'),
        Token(Position('file', 1, 8), TT.COLON),
        Token(Position('file', 2, 3), TT.STRING, 'euw'),
        Token(Position('file', 1, 14), TT.NEWLINE),
        Token(Position('file', 1, 14), TT.NEWLINE),
        Token(Position('file', 1, 14), TT.NEWLINE),
        Token(Position('file', 1, 1), TT.STRING, 'bucket'),
        Token(Position('file', 1, 1), TT.WHITESPACE),
        Token(Position('file', 1, 8), TT.STRING, 'nom'),
        Token(Position('file', 1, 8), TT.MINUS),
        Token(Position('file', 1, 8), TT.INT, 8),
        Token(Position('file', 1, 8), TT.COLON),
        Token(Position('file', 1, 14), TT.NEWLINE),
        Token(Position('file', 2, 1), TT.TAB),
        Token(Position('file', 2, 2), TT.STRING, 'region'),
        Token(Position('file', 1, 8), TT.COLON),
        Token(Position('file', 2, 3), TT.STRING, 'euw'),
        Token(Position('file', 1, 14), TT.NEWLINE),
        Token(Position('file', 2, 4), TT.EOF),
    ]
    parser = TokenParser(tokenized_input)
    output = parser.run()

    assert str(output) == '<RESOURCE \
type = <STRING-EXPR <STRING (STRING bucket)>>, \
name = <STRING-EXPR <STRING (STRING nom)> <STRING (STRING -)> <INT (INT 8)>>, \
parameters = <DICT <PARAMETER name = <STRING (STRING region)>, \
value = <STRING-EXPR <STRING (STRING euw)>>>>>\n\
<RESOURCE \
type = <STRING-EXPR <STRING (STRING bucket)>>, \
name = <STRING-EXPR <STRING (STRING nom)> <STRING (STRING -)> <INT (INT 8)>>, \
parameters = <DICT <PARAMETER name = <STRING (STRING region)>, \
value = <STRING-EXPR <STRING (STRING euw)>>>>>'



"""Lexer module for the DSL Parser."""
from thipster.engine.parsed_file import Position
from thipster.helpers import create_logger

from .exceptions import DSLParserNoEndingQuotesError
from .lexer_position import LexerPosition
from .token import TOKENTYPES as TT
from .token import Token


class Lexer:
    """Lexer for the DSL Parser."""

    def __init__(self, files: dict[str, str]):
        """Lexer for the DSL Parser.

        Parameters
        ----------
        files : dict[str, str]
            Dictionnary of files to tokenize, fileName : fileContent
        """
        self.__logger = create_logger(__name__)
        self.files = files
        self.tokenList = []
        self.__lexerPosition = LexerPosition()
        self.__currentChar = ''

    def run(self) -> list[Token]:
        """Launch the Lexer.

        Returns
        -------
        list[Token]
            List of Tokens representing the input files
        """
        self.__logger.info('Start Lexer')
        for filename, file_content in self.files.items():
            self.lex(filename, file_content)

        self.__logger.info(
            'Shuting down Lexer, processed %d file(s) and generated %d tokens',
            len(self.files),
            len(self.tokenList),
        )
        return self.tokenList

    def lex(
        self,
        filename: str,
        file_content: str,
    ) -> None:
        """Tokenize a file contents.

        Parameters
        ----------
        filename : str
            Name of the file
        file_content : str
            Content of the file

        Raises
        ------
        DSLParserNoEndingQuotes
            Syntax error : detected a starting quote but no ending one
        """
        self.__logger.debug('Lex file %s', filename)
        self.__lexerPosition = LexerPosition(filename)

        for char in file_content:
            self.__logger.debug('char %s', char)
            self.__currentChar = char

            if self.__lexerPosition.isComment:
                if char == '\n':
                    self.__lexerPosition.isComment = False
                    self.__handle_newline_token()
                continue

            if self.__lexerPosition.escaped:
                match char:
                    case '\n':
                        self.__lexerPosition.isMultiLine = True
                        self.__lexerPosition.new_line()
                    case _:
                        self.__add_next_char()
                self.__lexerPosition.escaped = False
                continue

            if not self.__lexerPosition.isQuotedString:
                self.__handle_syntax_tokens()
                continue

            if char in ['"', "'"]:
                self.__handle_quotes(char)()
                continue

            if char == '\n':
                if self.__lexerPosition.isMultiLine:
                    self.__lexerPosition.new_line()
                    continue
                position = str(
                    Position(
                        self.__lexerPosition.currentFile,
                        self.__lexerPosition.currentLine,
                        self.__lexerPosition.currentColumn,
                    ),
                )
                self.__logger.error(
                    f'Invalid syntax, missing ending quotes at : {position}',
                )
                raise DSLParserNoEndingQuotesError(position)

            self.__add_next_char()

        self.__end_file()
        self.__logger.debug('Finished lexing file %s', filename)

    def __add_next_char(self):
        """Add the current character to the current Token."""
        self.__lexerPosition.add_to_current_token(self.__currentChar)
        self.__lexerPosition.next_column()

    def __handle_empty_token(self):
        """Pass EMPTY tokens ''."""
        pass

    def __handle_simple_token(self, tokentype: TT):
        def handle_token():
            self.__handle_base_token(tokentype)
        return handle_token

    def __handle_word_token(self, tokentype: TT, value=None):
        def handle_token():
            """Handle an AMOUNT token 'amount'."""
            self.__lexerPosition.reset_consecutive_whitespaces()
            self.__add_base_token(tokentype, value, is_current_token=True)
        return handle_token

    def __handle_syntax_tokens(self) -> None:
        """Handle single character tokens.

        Check if the current character is a special token and calls the corresponding
        function.
        If it isn't, it calls the '__iterateNextChar' function
        """
        single_char_tokens = {
            ':': self.__handle_simple_token(TT.COLON),
            ',': self.__handle_simple_token(TT.COMMA),
            '[': self.__handle_simple_token(TT.BRACKETS_START),
            ']': self.__handle_simple_token(TT.BRACKETS_END),
            '(': self.__handle_simple_token(TT.PARENTHESES_START),
            ')': self.__handle_simple_token(TT.PARENTHESES_END),
            '"': self.__handle_quotes('"'),
            "'": self.__handle_quotes("'"),
            '#': self.__handle_hash_token,
            '\n': self.__handle_newline_token,
            '\\': self.__handle_backslash_token,
            '\t': self.__handle_simple_token(TT.TAB),
            ' ': self.__handle_whitespace,

            '=': self.__handle_simple_token(TT.EQ),
            '/': self.__handle_div_token,
            '-': self.__handle_simple_token(TT.MINUS),
            '+': self.__handle_simple_token(TT.PLUS),
            '*': self.__handle_simple_token(TT.MUL),
            '^': self.__handle_simple_token(TT.POW),
            '%': self.__handle_simple_token(TT.PERCENT),
            '<': self.__handle_simple_token(TT.LT),
            '>': self.__handle_simple_token(TT.GT),
            '!': self.__handle_simple_token(TT.EXCLAMATION),
        }

        single_char_tokens.get(self.__currentChar, self.__add_next_char)()

    def __handle_current_token(self) -> None:
        """Process the current stored token.

        Check if the current token is a keyword, calls the cprresponding function, or
        the '__handleLiteralsAndVariables' otherwise.
        """
        current_token = self.__lexerPosition.currentToken.lower()

        keywords = {
            '': self.__handle_empty_token,
            'amount': self.__handle_word_token(TT.AMOUNT),
            'and': self.__handle_word_token(TT.AND),
            'if': self.__handle_word_token(TT.IF),
            'elif': self.__handle_word_token(TT.ELIF),
            'else': self.__handle_word_token(TT.ELSE),
            'or': self.__handle_word_token(TT.OR),
            'true': self.__handle_word_token(TT.BOOLEAN, 'true'),
            'false': self.__handle_word_token(TT.BOOLEAN, 'false'),
            'not': self.__handle_word_token(TT.NOT),
        }

        keywords.get(current_token, self.__handle_literals_and_variables)()

    def __handle_literals_and_variables(self) -> None:
        """Create a literal or var token.

        Checks the current token to create a var, int, float or string token
        """
        current_token = self.__lexerPosition.currentToken
        if self.__lexerPosition.isVariable:
            self.__add_literal_token_to_list(TT.VAR, current_token)
            self.__lexerPosition.isVariable = False

        elif current_token.isdigit():
            self.__add_literal_token_to_list(TT.INT, current_token)

        elif self.__isfloat(current_token):
            self.__add_literal_token_to_list(TT.FLOAT, current_token)

        else:
            self.__add_literal_token_to_list(TT.STRING, current_token)

    def __add_literal_token_to_list(self, token_type: str, value: str):
        """Add a literal token to the token list.

        Parameters
        ----------
        tokenType : str
            Type of the token
        value : str
            Value of the token
        """
        self.__lexerPosition.reset_consecutive_whitespaces()
        self.__add_base_token(
            token_type,
            value,
            is_current_token=True,
        )

    def __add_base_token(
        self,
        token_type: str,
        value: str | None = None,
        is_current_token: bool = False,
    ) -> None:
        """Add a token to the token list.

        Parameters
        ----------
        tokenType : str
            Type of the token
        value : str, optional
            Value of the token, by default None
        isCurrentToken : bool, optional
            Indicates if this token is the current stored token, by default False
        """
        if not is_current_token:
            self.__add_token_to_list(
                Token(
                    self.__lexerPosition.currentCharPosition,
                    token_type,
                    value,
                ),
            )
            return

        self.__add_token_to_list(
            Token(
                self.__lexerPosition.currentTokenPosition,
                token_type,
                value,
            ),
        )

    def __update_position(
        self,
        reset_current_token: bool = True,
    ) -> None:
        """Update the lexer's current position.

        Parameters
        ----------
        resetCurrentToken : bool, optional
            Indicates if the current stored token should be reset, by default True
        """
        self.__lexerPosition.next_column()
        if reset_current_token:
            self.__lexerPosition.reset_current_token()

        self.__lexerPosition.reset_consecutive_whitespaces()
        self.__lexerPosition.isVariable = False

    def __handle_base_token(
        self,
        token_type: str,
        value: str | None = None,
        is_current_token: bool = False,
        handle_current_token: bool = True,
        reset_current_token: bool = True,
    ) -> None:
        """Handle a token creation and lexer position.

        Parameters
        ----------
        token_type : str
            Type of the token
        value : str, optional
            Value of the token, by default None
        is_current_token : bool, optional
            Indicates if this token is the current stored token, by default False
        handle_current_token : bool, optional
            Indicates if the current stored token should be processed, by default True
        reset_current_token : bool, optional
            Indicates if the current stored token should be reset, by default True
        """
        if handle_current_token:
            self.__handle_current_token()
        self.__add_base_token(token_type, value, is_current_token)
        self.__update_position(reset_current_token)

    def __handle_hash_token(self) -> None:
        """Handle a HASH token '#'.

        Sets a variable to indicate that the following literal is a Variable
        """
        self.__handle_current_token()
        self.__update_position()
        self.__lexerPosition.isVariable = True

    def __handle_div_token(self):
        if len(self.tokenList) == 0 or self.tokenList[-1].token_type is not TT.DIV:
            return self.__handle_simple_token(TT.DIV)()

        self.__lexerPosition.isComment = True
        self.__rm_last_tokens()
        return True

    def __handle_whitespace(self) -> None:
        """Handle a WHITESPACE token ' '.

        Replaces 4 following whitespaces by a TAB token
        """
        self.__handle_current_token()
        self.__add_base_token(TT.WHITESPACE)
        self.__lexerPosition.next_column()
        self.__lexerPosition.reset_current_token()
        self.__lexerPosition.isVariable = False
        self.__lexerPosition.increment_consecutive_whitespaces()
        if self.__lexerPosition.consecutiveWhitespaces == 4:
            self.__lexerPosition.next_column(-4)
            self.__rm_last_tokens(4)
            self.__add_base_token(TT.TAB)
            self.__lexerPosition.next_column(4)
            self.__lexerPosition.reset_consecutive_whitespaces()
            self.__lexerPosition.set_current_token_index()

    def __handle_newline_token(self) -> None:
        r"""Handle a NEWLINE token '\\n'."""
        if not self.__lexerPosition.isMultiLine:
            self.__handle_base_token(TT.NEWLINE)

        self.__lexerPosition.new_line()
        if not self.__lexerPosition.isMultiLine:
            self.__lexerPosition.set_current_token_index()

        self.__lexerPosition.isMultiLine = False

    def __handle_backslash_token(self) -> None:
        r"""Handle a BACKSLASH token '\\'.

        Sets a variable to indicate that the following line is part of the same token.
        """
        self.__lexerPosition.escaped = True
        self.__lexerPosition.next_column()

    def __handle_quotes(self, char) -> None:
        """Handle a DOUBLEQUOTES token '"'.

        Sets a variable to indicate that the following characters are a STRING token.
        """

        def quote_handler(self=self):
            if self.__lexerPosition.isQuotedString:
                if char == self.__lexerPosition.isQuotedString:
                    self.__add_literal_token_to_list(
                        TT.STRING, self.__lexerPosition.currentToken,
                    )
                    self.__lexerPosition.isQuotedString = False
                    self.__update_position()
                else:
                    self.__add_next_char()
            else:
                self.__handle_current_token()
                self.__lexerPosition.next_column()
                self.__lexerPosition.isQuotedString = char
                self.__lexerPosition.reset_current_token()

        return quote_handler

    def __handle_eof_token(self) -> None:
        """Add an EOF token at the end of each file."""
        self.__add_base_token(TT.EOF)

    def __add_token_to_list(self, token: Token | None) -> None:
        """Add a new token to the parser's token list.

        Parameters
        ----------
        token : Token | None
            New token to add
        """
        if not token:
            pass
        self.tokenList.append(token)

    def __rm_last_tokens(self, amount: int = 1):
        self.tokenList = self.tokenList[:-amount]

    def __end_file(self) -> None:
        """Add a NEWLINE and an EOF token at the end of each file."""
        if len(self.__lexerPosition.currentToken.strip()) > 0:
            self.__handle_current_token()
            self.__lexerPosition.reset_current_token()

        self.__lexerPosition.isMultiLine = False
        self.__handle_newline_token()
        self.__handle_eof_token()

    def __isfloat(self, num: str) -> bool:
        """Check if the string is a float.

        Parameters
        ----------
        num : str
            The input string to check

        Returns
        -------
        bool
            Is the string a float true or false
        """
        try:
            float(num)
            return True
        except ValueError:
            return False



"""THipster's Terraform CDK module."""
import copy
import importlib
import os
import shutil
import subprocess
import sys
import uuid
from pathlib import Path

from cdktf import App, TerraformStack
from constructs import Construct
from python_terraform import Terraform

import thipster.engine.parsed_file as pf
import thipster.engine.resource_model as rm
import thipster.terraform.exceptions as cdk_exceptions
from thipster.engine import AuthPort, TerraformPort
from thipster.helpers import create_logger


class ResourceCreationContext:
    """Context from which a resource is created."""

    def __init__(
        self,
        stack_self: TerraformStack,
        resource_name: str | None = None,
        resource_type: str | None = None,
        resource_class: type | None = None,
        parent_name: str | None = None,
        parent_type: str | None = None,
        parent_args: dict | None = {},
        arg_to_complete: str | None = None,
        no_modif: bool = True,
        no_dependencies: bool = False,
    ):
        """Context from which a resource is created.

        Parameters
        ----------
        stack_self : TerraformStack
            Stack in which the resource is created
        resource_name : str, optional
            Name of the resource, default None
        resource_type : str, optional
            Type of the resource, default None
        resource_class : type, optional
            Class of the resource (Python CDK), default None
        parent_name : str, optional
            Name of the parent resource, default None
        parent_type : str, optional
            Type of the parent resource, default None
        parent_args : dict, optional
            Arguments of the parent resource, default {}
        arg_to_complete : str, optional
            Name of the argument to complete, default None
        no_modif : bool, optional
            If the resource is created with default values, default True
        no_dependencies : bool, optional
            If the resource is created without dependencies, default False
        """
        self.stack_self = stack_self

        self.model: rm.ResourceModel = None
        self.dependencies = None
        self.__resource_type = None

        # Parent resource info
        self.parent_name = parent_name
        self.parent_type = parent_type
        self.parent_args = parent_args

        self.arg_to_complete = arg_to_complete
        self.no_modif: bool = no_modif
        self.no_dependencies: bool = no_dependencies

        # Resource info
        self.resource_name = resource_name
        self.resource_type = resource_type
        self.resource_class = resource_class
        self.resource_args = {}

    @classmethod
    def from_parent(cls, parent, **args):
        """Create context from a parent context.

        Parameters
        ----------
        ctx: ResourceCreationContext
            Context from which the resource is created
        """
        parent_name = parent.parent_name if parent.parent_name else parent.resource_name
        return ResourceCreationContext(
            parent.stack_self,
            resource_name=f'{parent_name}-{uuid.uuid4().hex[:8]}',
            parent_name=parent_name,
            parent_type=parent.resource_type,
            parent_args=parent.resource_args,
            **args,
        )

    def regenerate(self):
        """Regenerate resource name."""
        self.resource_name = f'{self.parent_name}-{uuid.uuid4().hex[:8]}'
        return self

    @property
    def resource_type(self):
        """Return resource type."""
        return self.__resource_type

    @resource_type.setter
    def resource_type(self, value):
        self.__resource_type = value

        if value is not None:
            self.model: rm.ResourceModel = CDK._models[self.resource_type]
            self.dependencies = copy.deepcopy(self.model.dependencies)
            if self.arg_to_complete in self.dependencies:
                del self.dependencies[self.arg_to_complete]


class CDK(TerraformPort):
    """Terraform code generation and usage.

    Works using the CDK for Terraform in python, and the python_terraform library.
    """

    _models = []
    _parent_resources_stack = []
    _resources_to_create: list[ResourceCreationContext] = []

    _inherited_attributes: list[pf.ParsedAttribute] = []
    _created_resources = {}
    _logger = create_logger(__name__)

    @classmethod
    def apply(cls, plan_file_path: str):
        """Apply generated Terraform plan.

        Parameters
        ----------
        plan_file_path : str
            Path to the plan file

        Returns
        -------
        str
            Terraform apply output
        """
        output = subprocess.run(
            ['terrraform', 'apply', plan_file_path],
            shell=True,
            capture_output=True,
            encoding='utf-8',
        )
        return output.stdout + output.stderr

    @classmethod
    def generate(
        cls, file: pf.ParsedFile, models: dict[str, rm.ResourceModel],
        _authenticator: AuthPort,
    ):
        """Generate Terraform file from given parsed file and models.

        Parameters
        ----------
        file : pf.ParsedFile
            ParsedFile object to transform
        models : dict[str, rm.ResourceModel]
            Associated models
        _auth : AuthPort
            Authenticator to use
        """
        cls._created_resources = {}
        cls._models = models
        cls._parent_resources_stack = []

        # Init CDK
        app = App()

        f_position = file.resources[0].position
        file_name = Path(f_position.filename).parts[-1] if f_position \
            else 'thipster_infrastructure'

        cls._logger.debug('Creating tf code for file %s', file_name)

        # Declare new stack in CDK

        class _ResourceStack(TerraformStack):
            def __init__(self, scope: Construct, ns: str):

                super().__init__(scope, ns)

                _authenticator.authenticate(self)

                for resource in file.resources:
                    ctx = ResourceCreationContext(self)
                    created_resource = _create_resource_from_resource(
                        ctx,
                        resource=resource,
                    )

                    cls._created_resources[f'{resource.resource_type}/{resource.name}']\
                        = created_resource

        _ResourceStack(app, file_name)

        # Synth files
        app.synth()

        output_directories = [
            f'{app.outdir}/stacks/{c.node.path}' for c in app.node.children
        ]

        CDK._logger.info(
            'Created %s terraform file(s)',
            len(output_directories),
        )

        # Move files
        for dirname in output_directories:
            shutil.move(
                Path(Path.cwd(), dirname, 'cdk.tf.json'),
                Path(Path.cwd(), 'thipster.tf.json'),
            )

        # Delete cdktf.out directory
            for content in os.listdir(Path(Path.cwd(), dirname)):
                if Path(f'{dirname}/{content}').is_dir():
                    shutil.rmtree(f'{dirname}/{content}')
                    continue

                Path(f'{dirname}/{content}').unlink()
            Path(dirname).rmdir()
        for content in os.listdir(Path(Path.cwd(), 'cdktf.out')):
            d = f'cdktf.out/{content}'
            Path(d).rmdir() if Path(d).is_dir() else Path(d).unlink()
        Path('cdktf.out').rmdir()

    @classmethod
    def init(cls):
        """Initilize terraform.

        Returns
        -------
        str
            Terraform init output
        """
        t = Terraform()
        _, stdout, stderr = t.init(upgrade=True)
        return stdout + stderr

    @classmethod
    def plan(cls, plan_file_path: str):
        """Get plan from generated terraform code.

        Parameters
        ----------
        plan_file_path : str
            Path and name of the plan file

        Returns
        -------
        str
            Terraform plan output
        """
        t = Terraform()
        _, stdout, stderr = t.plan(out=plan_file_path)
        return stdout + stderr

    @classmethod
    def _pip_install(cls, package: str):
        """Install a package if it wasn't already installed by thipster.

        Parameters
        ----------
        package : str
            Package to install
        """
        if package not in sys.modules:
            subprocess.check_call(
                [sys.executable, '-m', 'pip', 'install', '-qqq', package],
            )

    @classmethod
    def _import(cls, package_name: str, module_name: str, class_name: str) -> type:
        """Import a class from any package.

        Parameters
        ----------
        package_name : str
            Package that contains the class
        module_name : str
            Module that contains the class
        class_name : str
            Class to import

        Returns
        -------
        type :
            the imported class
        """
        module = importlib.import_module(f'{package_name}.{module_name}')
        return getattr(module, class_name)


def _create_default_resource(ctx: ResourceCreationContext):
    """Create a resource with all default values.

    Parameters
    ----------
    ctx: ResourceCreationContext
        Context from which the resource is created
    """
    # Checks for cyclic dependency
    if ctx.resource_type in CDK._parent_resources_stack:
        CDK._logger.error(
            '{ctx.resource_type} already present in parent Stack',
        )
        raise cdk_exceptions.CDKCyclicDependenciesError(
            CDK._parent_resources_stack,
        )

    CDK._parent_resources_stack.append(ctx.resource_type)

    attributes = copy.deepcopy(ctx.model.attributes)

    for a in CDK._inherited_attributes:
        if a.name in attributes:
            attributes[a.name].default = rm.ModelLiteral(a.value)

        else:
            attributes[a.name] = rm.ModelAttribute(
                cdk_name=a.name,
                default=rm.ModelLiteral(a.value),
            )

    # Checks that all attributes have a default value
    if ctx.no_modif and not all(
        x.optional or (x.default is not None) for x in attributes.values()
    ):
        missing_atributes = [
            k for k, v in attributes.items() if not v.optional and v.default is None
        ]

        raise cdk_exceptions.CDKMissingAttributeInDependencyError(
            ctx.resource_type,
            missing_atributes,
        )

    for attribute_name, attribute_value in attributes.items():
        _process_attribute(
            ctx, pf.ParsedAttribute(
                name=attribute_name,
                position=None,
                value=(
                    attribute_value.default
                    if isinstance(attribute_value.default, pf.ParsedLiteral)
                    else pf.ParsedLiteral(attribute_value.default)
                ),
            ),
        )

    # Create default defendencies if needed
    if not ctx.no_dependencies:
        _generate_default_dependencies(ctx)

    if ctx.model.name_key:
        ctx.resource_args[ctx.model.name_key] = ctx.resource_name

    # Import package and class
    CDK._pip_install(ctx.model.cdk_provider)
    ctx.resource_class = CDK._import(
        ctx.model.cdk_provider, ctx.model.cdk_module, ctx.model.cdk_name,
    )

    CDK._logger.debug(
        f'Created default {ctx.resource_class} named {ctx.resource_name}',
    )


def _create_resource(ctx: ResourceCreationContext, resource_args: object):
    """Create a resource with the given values.

    Parameters
    ----------
    ctx: ResourceCreationContext
        Context from which the resource is created
    resource_args : object
        Data source to create the resource

    Returns
    -------
    object :
        Created resource
    """
    if isinstance(resource_args, pf.ParsedResource):
        return _create_resource_from_resource(
            ctx, resource_args,
        )

    if isinstance(resource_args, list):
        return _create_resource_from_args(
            ctx, resource_args,
        )
    return None


def _create_resource_from_args(
    ctx: ResourceCreationContext, input_args: list[pf.ParsedAttribute] | None,
):
    """Create a resource from a list of ParsedAttributes.

    Parameters
    ----------
    ctx: ResourceCreationContext
        Context from which the resource is created
    input_args : list[ParsedAttribute]
        data source to create the resource

    Returns
    -------
    object :
        Created resource
    """
    ctx.no_modif = False
    ctx.no_dependencies = True

    # Create resource with default values
    _create_default_resource(ctx)
    ctx.no_modif = True
    ctx.no_dependencies = False

    # Process attributes
    def attributes(attribute_list: list[pf.ParsedAttribute]):
        for attribute in attribute_list:
            if attribute.name == ctx.model.name_key:
                ctx.resource_name = attribute.value
                ctx.resource_args[ctx.model.name_key] = attribute.value
            else:
                _process_attribute(ctx, attribute)

    attributes(input_args)
    attributes(CDK._inherited_attributes)

    CDK._inherited_attributes += input_args
    _generate_default_dependencies(ctx)
    CDK._inherited_attributes = CDK._inherited_attributes[:-len(input_args)]

    return _instantiate_class(ctx)


def _create_resource_from_resource(
        ctx: ResourceCreationContext, resource: pf.ParsedResource,
):
    """Create a resource from a list of ParsedAttributes.

    Parameters
    ----------
    ctx: ResourceCreationContext
        Context from which the resource is created
    resource : ParsedResource
        Data source to create the resource

    Returns
    -------
    object :
        Created resource
    """
    ctx.resource_name = resource.name
    ctx.resource_type = resource.resource_type

    return _create_resource_from_args(ctx, resource.attributes)


def _instantiate_class(ctx: ResourceCreationContext):
    """Instantiate a class.

    Parameters
    ----------
    ctx: ResourceCreationContext
        Context from which the resource is created

    Returns
    -------
    object :
        Created resource
    """
    CDK._parent_resources_stack.remove(ctx.resource_type)

    if not ctx.arg_to_complete:

        # check args
        if ctx.no_modif and not all(
            value.get('optional') or ctx.resource_args.get(name)
            for name, value in ctx.model.internal_objects.items()
        ):
            missing_internal_objects = [
                name for name, value in ctx.model.internal_objects.items()
                if not value.get('optional') and not ctx.resource_args.get(name)
            ]

            raise cdk_exceptions.CDKMissingAttributeInDependencyError(
                ctx.resource_type,
                missing_internal_objects,
            )

        if not ctx.model.name_key:
            class_ = ctx.resource_class(**ctx.resource_args)
        else:
            class_ = ctx.resource_class(
                ctx.stack_self, ctx.resource_name, **ctx.resource_args,
            )

        while len(CDK._resources_to_create) != 0:
            to_create = CDK._resources_to_create.pop()
            to_create.resource_args[to_create.arg_to_complete] = class_.id

            created_resource = to_create.resource_class(
                to_create.stack_self, to_create.resource_name,
                **to_create.resource_args,
            )

            CDK._created_resources[f'{to_create.resource_type}/{to_create.resource_name}']\
                = created_resource

        CDK._logger.debug(
            f'Created {ctx.resource_class} named {ctx.resource_name}',
        )
        return class_

    CDK._resources_to_create.append(ctx)
    return True


def _process_attribute(ctx: ResourceCreationContext, attribute: pf.ParsedAttribute):
    """Process an attribute.

    Parameters
    ----------
    ctx: ResourceCreationContext
        Context from which the resource is created
    attribute : ParsedAttribute
        Attribute to process
    """
    if not ctx.no_dependencies and attribute.name in ctx.dependencies:
        _check_explicit_dependency(
            ctx, attribute.name, attribute.value,
        )
        del ctx.dependencies[attribute.name]
        return

    # Checks if attribute is an internal object
    if attribute.name in ctx.model.internal_objects:
        _create_internal_object(
            ctx,
            attribute.name,
            attribute.value,
        )
        return

    # Checks if attribute is expected as an attibute
    if attribute.name not in ctx.model.attributes:
        return

    # Processes list attribute
    attribute_value = attribute.unmarshalled_value

    # Sets attribute value
    if attribute_value is not None:
        ctx.resource_args[ctx.model.attributes[attribute.name].cdk_name]\
            = attribute_value


def _generate_default_dependencies(ctx: ResourceCreationContext):
    """Generate default dependencies and internal objects in a resource.

    Parameters
    ----------
    ctx: ResourceCreationContext
        Context from which the resource is created
    """
    # Create missing dependencies
    for dependency_name, dependency_object in ctx.dependencies.items():
        dep_ctx = ResourceCreationContext.from_parent(
            ctx,
            resource_type=dependency_object['resource'],
        )
        ctx.resource_args[dependency_name] = _create_dependency(dep_ctx)

    # Create default internal objects if needed
    for internal_object_name, internal_object in ctx.model.internal_objects.items():
        if internal_object_name not in ctx.resource_args:
            for default_args_json in internal_object['defaults']:

                # Transform in list of ParsedAttributes
                default_args = []
                for arg_key, arg_value in default_args_json.items():
                    default_args.append(
                        pf.ParsedAttribute(
                            arg_key, None, pf.ParsedLiteral(arg_value),
                        ),
                    )

                _create_internal_object(
                    ctx,
                    internal_object_name,
                    default_args,
                )


def _create_internal_object(
    ctx: ResourceCreationContext,
    internal_object_name: str,
    internal_object_args,
):
    """Create an internal object in a resource.

    Parameters
    ----------
    ctx: ResourceCreationContext
        Context from which the resource is created
    internal_object_name : str
        Name of the internal object
    internal_object_args : object
        Data source to create the internal object
    """
    internal_object_model = ctx.model.internal_objects[internal_object_name]

    internal_object_base_ctx = ResourceCreationContext.from_parent(
        ctx,
        resource_type=internal_object_model['resource'],
        arg_to_complete=internal_object_model.get('arg_to_complete'),
    )

    internal_object_type = internal_object_model['var_type'] \
        if 'var_type' in internal_object_model else 'Unknown'

    if internal_object_type.startswith('list') or internal_object_model['is_list']:
        if internal_object_name not in ctx.resource_args\
                and not internal_object_base_ctx.arg_to_complete:
            ctx.resource_args[internal_object_name] = []

        if (
            isinstance(internal_object_args, list)
            and len(internal_object_args) > 0
            and not isinstance(internal_object_args[0], pf.ParsedAttribute)
        ):
            for internal_object in internal_object_args:
                internal_object_ctx = copy.deepcopy(internal_object_base_ctx)
                internal_object_ctx.regenerate()

                if isinstance(internal_object, pf.ParsedDict):
                    res = _create_resource(
                        internal_object_ctx,
                        internal_object.value,
                    )
                    if not internal_object_base_ctx.arg_to_complete:
                        ctx.resource_args[internal_object_name] += [res]

                elif isinstance(internal_object, pf.ParsedLiteral):
                    _check_explicit_dependency(
                        internal_object_ctx,
                        internal_object_name,
                        internal_object.value,
                    )

            return True

        res = _create_resource(
            internal_object_base_ctx,
            internal_object_args,
        )
        if not internal_object_base_ctx.arg_to_complete:
            ctx.resource_args[internal_object_name] += [res]
        return True

    res = _create_resource(
        internal_object_base_ctx,
        internal_object_args,
    )
    if not internal_object_base_ctx.arg_to_complete:
        ctx.resource_args[internal_object_name] = res

    return True


def _create_dependency(
        ctx: ResourceCreationContext,
        dependency_attributes: list[pf.ParsedAttribute] = None,
):
    """Create a dependency in a resource.

    Parameters
    ----------
    ctx: ResourceCreationContext
        Context from which the resource is created
    dependency_attributes: list[pf.ParsedAttribute]
        list of attributes to pass to the dependency creation. Defaults to None
    """
    if dependency_attributes:
        ctx.no_modif = False

    _create_default_resource(ctx)

    ctx.no_modif = True

    # Process attributes
    def attributes(attribute_list: list[pf.ParsedAttribute]):
        for attribute in attribute_list:
            if attribute.name == ctx.model.name_key:
                ctx.resource_args[ctx.model.name_key] = attribute.value
            else:
                _process_attribute(ctx, attribute)

    if dependency_attributes:
        attributes(dependency_attributes)
    attributes(CDK._inherited_attributes)

    dependency = _instantiate_class(ctx)
    return dependency.id


def _check_explicit_dependency(
    ctx: ResourceCreationContext, attribute_name: str, attribute_value: str | dict,
):
    """Check if a dependency attribute was explicited before.

    If it wasn't : create a default dependency.

    Parameters
    ----------
    ctx: ResourceCreationContext
        Context from which the resource is created
    attribute_name: str
        Name of the attribute to check
    attribute_value: str | dict
        Value of the attribute to check
    """
    resource_attribute = 'id'
    explicit_value = attribute_value
    if isinstance(attribute_value, pf.ParsedLiteral):
        explicit_value = attribute_value.value

    if isinstance(explicit_value, str):
        split = explicit_value.split('.', maxsplit=1)
        explicit_value = split[0]
        resource_attribute = split[1] if len(split) > 1 else 'id'

    dependency_type = ctx.dependencies[attribute_name]['resource']
    created_name = f'{dependency_type}/{explicit_value}'

    # Checks if attribute is an explicit dependency
    if created_name not in CDK._created_resources.keys():

        if isinstance(explicit_value, str):
            raise cdk_exceptions.CDKDependencyNotDeclaredError(
                attribute_name, attribute_value,
            )

        dep_ctx = ResourceCreationContext.from_parent(
            ctx,
            resource_type=dependency_type,
        )

        # Creates explicit dependency
        if (
            isinstance(attribute_value, list)
            and len(attribute_value) > 0
            and isinstance(attribute_value[0], pf.ParsedDict)
        ):
            ctx.resource_args[attribute_name] = []
            for dict_value in attribute_value:
                dep_ctx.regenerate()
                ctx.resource_args[attribute_name].append(
                    _create_dependency(
                        dep_ctx, dict_value.value,
                    ),
                )
            return True

        ctx.resource_args[attribute_name] = _create_dependency(
            dep_ctx, attribute_value,
        )
        return True

    resource = getattr(
        CDK._created_resources[created_name], resource_attribute,
    )
    if attribute_name in ctx.resource_args:
        ctx.resource_args[attribute_name] += [resource]

    ctx.resource_args[attribute_name] = resource
    return True



"""Basic tests for the engine module functions."""

import pytest

import thipster.engine as eng
from thipster.engine.parsed_file import ParsedFile
from thipster.engine.resource_model import ResourceModel

test_file = 'test.file'


def logger(name: str):
    """Log the function calls in the console."""
    def wrapper(function):
        def internal_wrapper(*args, **kwargs):
            print(f'{name} starting')
            res = function(*args, **kwargs)
            print(f'{name} returned :\n{res!s}')
            return res
        return internal_wrapper
    return wrapper


class MockError(Exception):
    """Mock errors for testing purposes."""

    def __init__(self, message):
        self.message = message


class MockAuth(eng.AuthPort):
    """Mock engine authentification port."""

    @logger('- Authentifier')
    def authenticate(self):
        """Mock authentification method."""
        return (None, None)


class MockParser(eng.ParserPort):
    """Mock engine parser port."""

    @logger('- Parser')
    def run(self, filename) -> str:  # noqa: ARG002
        """Mock parser method."""
        return ParsedFile()


class MockRepository(eng.RepositoryPort):
    """Mock engine repository port."""

    @logger('- Repo')
    def get(
        self,
        resource_names: list[str],  # noqa: ARG002
    ) -> dict[str, ResourceModel]:
        """Mock the get models repository method."""
        return {}


class MockTerraform(eng.TerraformPort):
    """Mock engine terraform port."""

    @logger('- Terraform:apply')
    def apply(self, file):
        """Mock the terraform apply method."""
        pass

    @logger('- Terraform:generate')
    def generate(self, a, b, auth):
        """Mock the terraform generate method."""
        pass

    @logger('- Terraform:init')
    def init(self):
        """Mock the terraform init method."""
        pass

    @logger('- Terraform:plan')
    def plan(self, file):
        """Mock the terraform plan method."""
        pass


def test_engine_calls():
    """Test the engine calls."""
    parser = MockParser()
    repository = MockRepository()
    auth = MockAuth()
    terraform = MockTerraform()

    engine = eng.Engine(parser, repository, auth, terraform)
    res = engine.run(test_file)

    assert res is None


def test_parser_failure(mocker):
    """Test the parser failure."""

    def parser_fail(self, filename: str):  # noqa: ARG001
        msg = 'Parser failure'
        raise MockError(msg)

    mocker.patch(
        'tests.engine.test_engine.MockParser.run',
        parser_fail,
    )

    parser = MockParser()
    repository = MockRepository()
    auth = MockAuth()
    terraform = MockTerraform()

    with pytest.raises(MockError):
        engine = eng.Engine(parser, repository, auth, terraform)
        engine.run(test_file)


def test_repository_failure(mocker):
    """Test the behavior of the repository when it fails."""

    def repository_fail(
        self,  # noqa: ARG001
        resource_names: list[str],  # noqa: ARG001
    ) -> dict[str, ResourceModel]:
        msg = 'Repository failure'
        raise MockError(msg)

    mocker.patch(
        'tests.engine.test_engine.MockRepository.get',
        repository_fail,
    )

    parser = MockParser()
    repository = MockRepository()
    auth = MockAuth()
    terraform = MockTerraform()

    with pytest.raises(MockError):
        engine = eng.Engine(parser, repository, auth, terraform)
        engine.run(test_file)



"""ParserFactory module."""
import os
from pathlib import Path

from thipster.engine import ParserPort
from thipster.engine.parsed_file import ParsedFile

from .dsl_parser import DSLParser
from .exceptions import (
    NoFileFoundError,
    ParserPathNotFoundError,
)
from .yaml_parser import YAMLParser


class NoParser(ParserPort):
    """Used when no parser is found for a file extension."""

    @classmethod
    def run(cls, path) -> ParsedFile:  # noqa: ARG003
        """Run the Parser."""
        return ParsedFile()


class ParserFactory(ParserPort):
    """Parser factory, used to run the right parser on the right file."""

    __parsers = {
        '.yaml': YAMLParser,
        '.yml': YAMLParser,
        '.jinja': YAMLParser,
        '.thips': DSLParser,
    }

    @classmethod
    def add_parser(cls, parser: ParserPort, extensions: list[str]):
        """Add a parser to the ParserFactory."""
        cls.__parsers.update({e: parser for e in extensions})

    @classmethod
    def __getfiles(cls, path: str) -> list[str]:
        """Get the file(s) on the requested path.

        Recursively get all files names in the requested directory and its
        sudirectories. Can be run on a path file as well.

        Parameters
        ----------
        path: str
            Path to run this function onto

        Returns
        -------
        list[str]
            A list of all the filenames
        """
        path = Path(path).resolve().as_posix()

        if not Path(path).exists():
            raise ParserPathNotFoundError(path)

        files = []

        if Path(path).is_dir():
            for content in os.listdir(path):
                files += cls.__getfiles(f'{path}/{content}')

        if Path(path).is_file():
            return [path]

        return files

    @classmethod
    def run(cls, path: str) -> ParsedFile:
        """Run the ParserFactory.

        Parameters
        ----------
        path: str
            Path to run the parser into

        Returns
        -------
        ParsedFile
            A ParsedFile object with the content of all the files in the input path
        """
        files = cls.__getfiles(path)

        res = ParsedFile()
        for file in files:
            parsed_file = cls.__get_parser(file).run(file)
            res.resources += parsed_file.resources

        if len(res.resources) == 0:
            raise NoFileFoundError(path)

        return res

    @classmethod
    def __get_parser(cls, path) -> ParserPort:

        path_extension = Path(path).suffix

        return cls.__parsers.get(
            path_extension, NoParser,
        )



"""ResourceModel.py module."""

from abc import ABC


class ModelValue(ABC):
    """Model Attribute Value Interface."""

    value = None


class ModelAttribute:
    """Represents a Resource Model attribute."""

    def __init__(
            self, cdk_name: str,
            default: ModelValue | None = None,
            optional: bool = True,
            is_list: bool = False,
    ):
        """Represent a Resource Model attribute.

        Parameters
        ----------
        name : str
            Attribute name
        default : ModelValue, optional
            Default Attribute value if there is one, by default None
        optional : bool, optional
            Is attribute optional ?, by default True
        is_list : bool, optional
            Is attribute a list ?, by default False
        """
        self.cdk_name = cdk_name
        self.__default = default
        self.optional = optional
        self.is_list = is_list

    @property
    def default(self):
        """Default value of the attribute."""
        return self.__default.value if self.__default else None

    @default.setter
    def default(self, value):
        self.__default = value


class ModelList(ModelValue):
    """Represent a List of values for a Resource Model attribute."""

    def __init__(self, value: list[ModelValue | None] | None):
        """Represent a List of values for a Resource Model attribute.

        Parameters
        ----------
        value : list[ModelValue | None] | None
            List of values
        """
        super().__init__()
        self.value = value

    def __iter__(self):
        """Iterate over the list."""
        self.i = 0
        return self

    def __next__(self):
        """Return the next value of the iterator."""
        if not self.value:
            return None
        ret = self.value[self.i]
        if self.i > len(self.value):
            raise StopIteration

        self.i += 1

        return ret


class ModelLiteral(ModelValue):
    """Represent a literal value for a Resource Model attribute."""

    def __init__(self, value):
        """Represent a literal value for a Resource Model attribute.

        Parameters
        ----------
        value : bool | int | float | str
            Literal value
        """
        super().__init__()
        self.value = value


class ModelDict(ModelValue):
    """Represent a dictionary value for a Resource Model attribute."""

    def __init__(self, value: dict[str, ModelAttribute] | None):
        """Represent a dictionary value for a Resource Model attribute.

        Parameters
        ----------
        value : dict[str, ModelAttribute] | None
            Dictionary value
        """
        super().__init__()
        self.value = value

    def __iter__(self):
        """Iterate over the dictionary."""
        self.i = 0
        return self

    def __next__(self):
        """Return the next value of the iterator."""
        if not self.value:
            return None
        ret = self.value[self.i]
        if self.i > len(self.value):
            raise StopIteration

        self.i += 1

        return ret


class ResourceModel:
    """Represents a Resource Model."""

    def __init__(
            self,
            resource_type: str,
            attributes: dict[str, ModelAttribute] | None,
            dependencies: dict[str, dict[str, object]] | None,
            internal_objects: dict[str, dict[str, object]] | None,
            name_key: str | None,
            cdk_provider: str,
            cdk_module: str,
            cdk_name: str,
    ):
        """Represent a Resource Model.

        Parameters
        ----------
        resource_type : str
            Resource type
        attributes : dict[str, ModelAttribute] | None
            Resource attributes
        dependencies : dict[str, dict[str, object]] | None
            Resource dependencies
        internal_objects : dict[str, dict[str, object]] | None
            Resource internal objects
        name_key : str | None
            Key to declare the resource 'name' (if any) in the Terraform Python CDK
        cdk_provider : str
            Terraform Python CDK provider
        cdk_module : str
            Terraform Python CDK module
        cdk_name : str
            Resource name in the Terraform Python CDK
        """
        self.resource_type = resource_type
        self.attributes = attributes
        self.dependencies = dependencies
        self.internal_objects = internal_objects
        self.name_key = name_key
        self.cdk_provider = cdk_provider
        self.cdk_module = cdk_module
        self.cdk_name = cdk_name



"""Test the creation and string representation of a token."""
from thipster.engine.parsed_file import Position
from thipster.parser.dsl_parser.token import Token


def test_create_token():
    """Test the creation of a token."""
    position = Position(filename='testFile', ln=2, col=6)
    token_type = 'test_type'
    value = 'test_value'

    token = Token(position, token_type=token_type, value=value)

    assert isinstance(token, Token)
    assert token.position == position
    assert token.token_type == token_type
    assert token.value == value


def test_token_to_string():
    """Test the string representation of a token."""
    position = Position(filename='testFile', ln=2, col=6)
    position_str = '(File : testFile, Ln 2, Col 6)'
    assert str(position) == position_str

    token_type = 'test_type'
    value = 'test_value'
    token = Token(position, token_type=token_type, value=value)
    token_str = f'(Type: {token_type!s}, Position: {position_str}, Value: {value})'
    assert repr(token) == token_str



"""Exceptions for the repository module."""
from thipster.engine import THipsterError


class ModelNotFoundError(THipsterError):
    """Exception raised when a model is not found in the database."""

    def __init__(self, model: str, *args) -> None:
        """Exception raised when a model is not found in the database.

        Parameters
        ----------
        model: str
            The json resource model that was not found.
        """
        super().__init__(*args)
        self.model = model

    @property
    def message(self):
        """Return the exception message."""
        return f"Couldn't find {self.model} model"
